Attribute VB_Name = "modDelaunay3D"
'=================================================================
' Descrizione.....: Routines per la triangolazione Delaunay in 3D.
' Nome dei Files..: modDelaunay3D.bas, modUtilita
' Data............: 09/12/2003
' Versione........: 1.0 a 32 bits.
' Sistema.........: VB6 (SP5) sotto Windows NT (SP6).
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
' DownLoads a.....: http://www.flanguasco.org
'                   http://members.xoom.virgilio.it/flanguasco/
'=================================================================
'
'   Routines di ingresso:
'    DTRIS3 NPT, Sizht, Maxbf, Maxfc, Vcl(), VM(), _
            Nbf, Nfc, Nface, Ntetra, BF(), FC(), HT(), IERR
'    DTRIW3 NPT, Sizht, Maxbf, Maxfc, Vcl(), VM(), _
            Nbf, Nfc, Nface, Ntetra, BF(), FC(), HT(), IERR
'
'    Parametri: vedi descrizione all' interno delle routines.
'
'    Nota:  Tutti i vettori e le matrici di queste routines
'           iniziano dall' indice 1 (HT() escluso).
'
'    Nota:  Tutte le routines di questo modulo sono state tradotte
'           dal FORTRAN della libreria Geompack3.
'
Option Explicit
'
Private Const Epsilon# = 2# ^ -52 ' Epsilon del FORTRAN per numeri in doppia precisione.
Public Sub DTRIW3(ByVal NPT&, ByVal Sizht&, ByVal Maxbf&, ByVal Maxfc&, _
    Vcl#(), ByRef VM&(), ByRef Nbf&, ByRef Nfc&, ByRef Nface&, ByRef Ntetra&, _
    ByRef BF&(), ByRef FC&(), ByRef HT&(), ByRef IERR&)
'
'******************************************************************************
'
'  DTRIW3 constructs a Delaunay triangulation of vertices in 3D.
'
'
'  Purpose:
'
'    Construct Delaunay triangulation of 3D vertices using
'    incremental approach and local transformations. Vertices are
'    inserted one at a time in order given by VM array. The initial
'    tetrahedra created due to a new vertex are obtained by a walk
'    through the triangulation until location of vertex is known.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, NPT - number of 3D vertices (points).
'
'    Input, SIZHT - size of hash table HT; a good choice is a prime number
'    which is about 1/8 * NFACE (or 3/2 * NPT for random
'    points from the uniform distribution).
'
'    Input, MAXBF - maximum size available for BF array.
'
'    Input, MAXFC - maximum size available for FC array.
'
'    Input, VCL(1:3,1:*) - vertex coordinate list.
'
'    Input/output, VM(1:NPT) - indices of vertices of VCL being triangulated;
'    vertices are inserted in order given by VM, except that on output,
'    the third and fourth elements may be swapped so
'    that first 4 vertices are non-coplanar.
'
'    Output, NBF - number of positions used in BF array; NBF <= MAXBF.
'
'    Output, NFC - number of positions used in FC array; NFC <= MAXFC.
'
'    Output, NFACE - number of faces in triangulation; NFACE <= NFC.
'
'    Output, NTETRA - number of tetrahedra in triangulation.
'
'    Output, BF(1:3,1:NBF) -  array of boundary face records containing pointers
'    (indices) to FC; if FC(5,I) = -J < 0 and FC(1:3,I) = ABC,
'    then BF(1,J) points to other boundary face with edge BC,
'    BF(2,J) points to other boundary face with edge AC, and
'    BF(3,J) points to other boundary face with edge AB;
'    if BF(1,J) <= 0, record is not used and is in avail list.
'
'    Output, FC(1:7,1:NFC) - array of face records which are in linked lists
'    in hash table with direct chaining. Fields are:
'    FC(1:3,*) - A,B,C with 1<=A<B<C<=NPT; indices in VM of 3
'    vertices of face; if A <= 0, record is not used (it is
'    in linked list of avail records with indices <= NFC);
'    internal use: if B <= 0, face in queue, not in triang
'    FC(4:5,*) - D,E; indices in VM of 4th vertex of 1 or 2
'    tetrahedra with face ABC; if ABC is boundary face
'    then E < 0 and |E| is an index of BF array
'    FC(6,*) - HTLINK; pointer (index in FC) of next element
'    in linked list (or NULL = 0)
'    FC(7,*) - used internally for QLINK (link for queues or
'    stacks); pointer (index in FC) of next face in queue/
'    stack (or NULL = 0); QLINK = -1 indicates face is not
'    in any queue/stack, and is output value (for records
'    not in avail list), except:
'    FC(7,1:2) - HDAVBF,HDAVFC : head pointers of avail list in BF, FC.
'
'    Output, HT(0:SIZHT-1) - hash table using direct chaining; entries are
'    head pointers of linked lists (indices of FC array)
'    containing the faces and tetrahedra of triangulation
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
    Dim I&, I3&, I4&, Ifac&, Ind&, Ivrt&, Ptr&
    Dim Back&, Front&, hdavbf&, hdavfc&, vi
    Dim Ctr#(1 To 3)
    Const msglvl& = 0
    Dim PT1#(1 To 3) ' Vettore d' appoggio.
'
'  Find initial valid tetrahedron, and initialize data structures.
'
    IERR = 0
'
    Call FRSTET(False, NPT, Vcl(), VM(), I3, I4, IERR)
    If (IERR <> 0) Then Exit Sub
'
    For I = 1 To 3
        Ctr(I) = (Vcl(I, VM(1)) + Vcl(I, VM(2)) + Vcl(I, VM(3)) _
               + Vcl(I, VM(4))) / 4#
    Next I
'
    'ht(0:sizht-1) = 0
    hdavbf = 0
    hdavfc = 0
    Nbf = 4
    Nfc = 4
    Ntetra = 1
    Call HTINS(1, 1, 2, 3, 4, -1, NPT, Sizht, FC(), HT())
    Call HTINS(2, 1, 2, 4, 3, -2, NPT, Sizht, FC(), HT())
    Call HTINS(3, 1, 3, 4, 2, -3, NPT, Sizht, FC(), HT())
    Call HTINS(4, 2, 3, 4, 1, -4, NPT, Sizht, FC(), HT())
    BF(1, 1) = 4
    BF(2, 1) = 3
    BF(3, 1) = 2
    BF(1, 2) = 4
    BF(2, 2) = 3
    BF(3, 2) = 1
    BF(1, 3) = 4
    BF(2, 3) = 2
    BF(3, 3) = 1
    BF(1, 4) = 3
    BF(2, 4) = 2
    BF(3, 4) = 1
    Ifac = 4
'
    If (msglvl = 4) Then
        'write ( *,600) (vm(i),i=1,4),i3,i4
    End If
'
    ' Insert Ith vertex into Delaunay triang of first I-1 vertices.
    ' Walk through triang to find location of vertex I, create new
    ' tetrahedra, apply local transf based on empty sphere criterion.
    '
    For I = 5 To NPT
        vi = VM(I)
'
        If (msglvl = 4) Then
            'write ( *,610) i,vi
        End If
'
        If (FC(5, Ifac) = I - 1) Then
            Ivrt = 5
        Else
            Ivrt = 4
        End If
'
        'Call WALKT3(Vcl(1, vi), NPT, Sizht, Ntetra, Vcl, VM, FC, HT, Ifac, Ivrt, IERR)
        PT1(1) = Vcl(1, vi): PT1(2) = Vcl(2, vi): PT1(3) = Vcl(3, vi)
        Call WALKT3(PT1(), NPT, Sizht, Ntetra, Vcl(), VM(), FC(), HT(), Ifac, Ivrt, IERR)
        If (IERR <> 0) Then Exit Sub
'
        If (Ivrt = 6) Then
            Call NWTHFC(I, Ifac, NPT, Sizht, Nbf, Nfc, Maxbf, Maxfc, BF(), FC(), HT(), _
            Ntetra, hdavbf, hdavfc, Front, Back, IERR)
'
        ElseIf (Ivrt >= 4) Then
            Call NWTHIN(I, Ifac, Ivrt, NPT, Sizht, Nfc, Maxfc, FC(), HT(), Ntetra, _
                        hdavfc, Front, Back, IERR)
'
        ElseIf (Ivrt = 0) Then
            Front = Ifac
            'Call VBFAC(Vcl(1, vi), ctr, Vcl, VM, BF, FC, front, 0)
            PT1(1) = Vcl(1, vi): PT1(2) = Vcl(2, vi): PT1(3) = Vcl(3, vi)
            Call VBFAC(PT1(), Ctr(), Vcl(), VM(), BF(), FC(), Front, 0)
'
            If (IERR <> 0) Then
                Exit Sub
            End If
'
            Call NWTHOU(I, NPT, Sizht, Nbf, Nfc, Maxbf, Maxfc, BF(), FC(), HT(), _
                        Ntetra, hdavbf, hdavfc, Front, Back, Ind, IERR)
'
        ElseIf (Ivrt >= 1) Then
            Call NWTHED(I, Ifac, Ivrt, NPT, Sizht, Nbf, Nfc, Maxbf, Maxfc, BF(), FC(), _
            HT(), Ntetra, hdavbf, hdavfc, Front, Back, IERR)
'
        Else
            IERR = 302
        End If
'
        If (IERR <> 0) Then Exit Sub
'
        Call SWAPES(False, I, NPT, Sizht, Nfc, Maxfc, Vcl(), VM(), BF(), FC(), HT(), _
                    Ntetra, hdavfc, Front, Back, Ind, IERR)
'
        If (IERR <> 0) Then Exit Sub
'
        If (Ind <> 0) Then Ifac = Ind
    Next I
'
    Nface = Nfc
    Ptr = hdavfc
'
    Do While (Ptr <> 0)
        Nface = Nface - 1
        Ptr = -FC(1, Ptr)
    Loop
'
    FC(7, 1) = hdavbf
    FC(7, 2) = hdavfc
'
'600 format (/1x,'dtriw3: first tetrahedron: ',4i7/4x,'i3, i4 =',2i7)
'610 format (/1x,'step',i7,':   vertex i =',i7)
'
'
'
End Sub
Public Sub DTRIS3(ByVal NPT&, ByVal Sizht&, ByVal Maxbf&, ByVal Maxfc&, _
    Vcl#(), ByRef VM&(), ByRef Nbf&, ByRef Nfc&, ByRef Nface&, _
    ByRef Ntetra&, ByRef BF&(), ByRef FC&(), ByRef HT&(), ByRef IERR&)
'
'******************************************************************************
'
'  DTRIS3 constructs a Delaunay triangulation of vertices in 3D.
'
'
'  Purpose:
'
'    Construct Delaunay triangulation of 3D vertices using
'    incremental approach and local transformations.  Vertices are
'    first sorted in lexicographically increasing (x,y,z) order, and
'    then are inserted one at a time from outside the convex hull.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, NPT - number of 3D vertices (points).
'
'    Input, SIZHT - size of hash table HT; a good choice is a prime number
'    which is about 1/8 * NFACE (or 3/2 * NPT for random
'    points from the uniform distribution).
'
'    Input, MAXBF - maximum size available for BF array.
'
'    Input, MAXFC - maximum size available for FC array.
'
'    Input, VCL(1:3,1:*) - vertex coordinate list.
'
'    Input/utput, VM(1:NPT) - indices of vertices of VCL being triangulated.
'    On output, indices are permuted, so that VCL(*,VM(1)), ... ,
'    VCL(*,VM(NPT)) are in lexicographic increasing order,
'    with possible slight reordering so first 4 vertices are
'    non-coplanar.
'
'    Output, NBF - number of positions used in BF array; NBF <= MAXBF.
'
'    Output, NFC - number of positions used in FC array; NFC <= MAXFC.
'
'    Output, NFACE - number of faces in triangulation; NFACE <= NFC.
'
'    Output, NTETRA - number of tetrahedra in triangulation.
'
'    Output, BF(1:3,1:NBF) -  array of boundary face records containing pointers
'    (indices) to FC; if FC(5,I) = -J < 0 and FC(1:3,I) = ABC,
'    then BF(1,J) points to other boundary face with edge BC,
'    BF(2,J) points to other boundary face with edge AC, and
'    BF(3,J) points to other boundary face with edge AB;
'    if BF(1,J) <= 0, record is not used and is in avail list.
'
'    Output, FC(1:7,1:NFC) - array of face records which are in linked lists
'    in hash table with direct chaining. Fields are:
'    FC(1:3,*) - A,B,C with 1<=A<B<C<=NPT; indices in VM of 3
'    vertices of face; if A <= 0, record is not used (it is
'    in linked list of avail records with indices <= NFC);
'    internal use: if B <= 0, face in queue, not in triangulation.
'    FC(4:5,*) - D,E; indices in VM of 4th vertex of 1 or 2
'    tetrahedra with face ABC; if ABC is boundary face
'    then E < 0 and |E| is an index of BF array
'    FC(6,*) - HTLINK; pointer (index in FC) of next element
'    in linked list (or NULL = 0)
'    FC(7,*) - used internally for QLINK (link for queues or
'    stacks); pointer (index in FC) of next face in queue/
'    stack (or NULL = 0); QLINK = -1 indicates face is not
'    in any queue/stack, and is output value (for records
'    not in avail list), except:
'    FC(7,1:2) - HDAVBF,HDAVFC : head pointers of avail list in BF, FC.
'
'    Output, HT(0:SIZHT-1) - hash table using direct chaining; entries are
'    head pointers of linked lists (indices of FC array)
'    containing the faces and tetrahedra of triangulation.
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
    Dim a&, b&, Back&, bfi&, e&, Front&, hdavbf&, hdavfc&, I&, I3&, I4&, Ip&
    Dim J&, K&, op&, Ptr&, Topnv&, Top&, va&, vb&, vc&, vi&
    Dim Ctr#(1 To 3)
    Const msglvl& = 0
    Dim PT1#(1 To 3), PT2#(1 To 3), PT3#(1 To 3), PT4#(1 To 3)  ' Vettori d' appoggio.
'
    ' Permute elements of VM so that vertices are in lexicographic
    ' order, and initialize data structures.
'
    IERR = 0
    
    Call DHPSRT(3, NPT, 3, Vcl(), VM())
'
    Call FRSTET(True, NPT, Vcl(), VM(), I3, I4, IERR)
    If (IERR <> 0) Then Exit Sub
'
    For I = 1 To 3
        Ctr(I) = (Vcl(I, VM(1)) + Vcl(I, VM(2)) + Vcl(I, VM(3)) + Vcl(I, VM(4))) / 4#
    Next I
'
    'ht(0:sizht-1) = 0
    For I = 0 To Sizht - 1
        HT(I) = 0
    Next I
    hdavbf = 0
    hdavfc = 0
    Nbf = 4
    Nfc = 4
    Ntetra = 1
    Call HTINS(1, 1, 2, 3, 4, -1, NPT, Sizht, FC(), HT())
    Call HTINS(2, 1, 2, 4, 3, -2, NPT, Sizht, FC(), HT())
    Call HTINS(3, 1, 3, 4, 2, -3, NPT, Sizht, FC(), HT())
    Call HTINS(4, 2, 3, 4, 1, -4, NPT, Sizht, FC(), HT())
    BF(1, 1) = 4
    BF(2, 1) = 3
    BF(3, 1) = 2
    BF(1, 2) = 4
    BF(2, 2) = 3
    BF(3, 2) = 1
    BF(1, 3) = 4
    BF(2, 3) = 2
    BF(3, 3) = 1
    BF(1, 4) = 3
    BF(2, 4) = 2
    BF(3, 4) = 1
'
    If (msglvl = 4) Then
        'write ( *,600) (vm(i),i=1,4),i3,i4
    End If
'
    ' Insert I-th vertex into Delaunay triangle of first I-1 vertices.
'
    For I = 5 To NPT
        vi = VM(I)
'
        If (msglvl = 4) Then
            'write ( *,610) i,vi
        End If
'
        Ip = I - 1
        If (I = 5) Then Ip = 2
        If (I = I3 + 2) Then Ip = 3
        If (I = I4 + 1) Then Ip = 4
'
        ' Form stacks of boundary faces involving vertex IP.
        ' TOP is for stack of boundary faces to be tested for visibility.
        ' FRONT is for stack of boundary faces visible from vertex I.
        ' TOPNV is for stack of boundary faces not visible from I.
'
        Front = 0
        Topnv = 0
'
        If (I = 5) Then
            Top = 4
            a = 3
            b = 2
            If (Ip = 2) Then a = 2
            If (Ip <= 3) Then b = 1
            FC(7, Top) = a
            FC(7, a) = b
            FC(7, b) = 0
'
        ElseIf (Ip = I - 1) Then
            Top = bfi
            FC(7, bfi) = 0
            b = FC(2, bfi)
            Ptr = BF(1, -FC(5, bfi))
'
30  'continue
'
            If (FC(1, Ptr) = b) Then
                b = FC(2, Ptr)
                J = 1
            Else
                b = FC(1, Ptr)
                J = 2
            End If
'
            FC(7, Ptr) = Top
            Top = Ptr
            Ptr = BF(J, -FC(5, Ptr))
            If (Ptr <> bfi) Then GoTo 30
'
        Else
            For K = 1 To Nbf
'
                If (BF(1, K) <= 0) Then
                    GoTo Cycle 'cycle
                End If
'
                For e = 1 To 3
                    Ptr = BF(e, K)
                    If (FC(1, Ptr) = Ip) Then
                        b = FC(2, Ptr)
                        J = 3
                        GoTo 60
                    ElseIf (FC(2, Ptr) = Ip) Then
                        b = FC(1, Ptr)
                        J = 3
                        GoTo 60
                    ElseIf (FC(3, Ptr) = Ip) Then
                        b = FC(1, Ptr)
                        J = 2
                        GoTo 60
                    End If
                Next e
Cycle:
            Next K
'
60  'continue
'
            bfi = Ptr
            Top = bfi
            FC(7, bfi) = 0
            Ptr = BF(J, -FC(5, bfi))
'
70  'continue
'
            If (FC(1, Ptr) = b) Then
                J = 1
                If (FC(2, Ptr) = Ip) Then
                    b = FC(3, Ptr)
                Else
                    b = FC(2, Ptr)
                End If
            ElseIf (FC(2, Ptr) = b) Then
                J = 2
                If (FC(1, Ptr) = Ip) Then
                    b = FC(3, Ptr)
                Else
                    b = FC(1, Ptr)
                End If
            Else
                J = 3
                If (FC(1, Ptr) = Ip) Then
                    b = FC(2, Ptr)
                Else
                    b = FC(1, Ptr)
                End If
            End If
'
            FC(7, Ptr) = Top
            Top = Ptr
            Ptr = BF(J, -FC(5, Ptr))
            If (Ptr <> bfi) Then GoTo 70
'
        End If
'
        ' Find a boundary face visible from vertex I.
'
80  'continue
    
        If (Top = 0) Then GoTo 110
'
        Ptr = Top
        Top = FC(7, Ptr)
        va = VM(FC(1, Ptr))
        vb = VM(FC(2, Ptr))
        vc = VM(FC(3, Ptr))
        'op = OPSIDE(Vcl(1, va), Vcl(1, vb), Vcl(1, vc), Ctr, Vcl(1, vi))
        PT1(1) = Vcl(1, va): PT1(2) = Vcl(2, va): PT1(3) = Vcl(3, va)
        PT2(1) = Vcl(1, vb): PT2(2) = Vcl(2, vb): PT2(3) = Vcl(3, vb)
        PT3(1) = Vcl(1, vc): PT3(2) = Vcl(2, vc): PT3(3) = Vcl(3, vc)
        PT4(1) = Vcl(1, vi): PT4(2) = Vcl(2, vi): PT4(3) = Vcl(3, vi)
        op = OPSIDE(PT1(), PT2(), PT3(), Ctr, PT4())
'
        If (op = 2) Then
            IERR = 301
            Exit Sub
        End If
'
        If (op = 1) Then
            Front = Ptr
90  'continue
            If (Top = 0) Then GoTo 110
            Ptr = Top
            Top = FC(7, Ptr)
            FC(7, Ptr) = -1
            GoTo 90
        Else
            FC(7, Ptr) = Topnv
            Topnv = Ptr
        End If
'
        GoTo 80
'
110 'continue
'
        If (Front = 0) Then
            IERR = 306
            Exit Sub
        End If
'
        ' Find remaining visible boundary faces, add new tetrahedra with
        ' vertex I, apply local transformation based on empty sphere criterion.
'
        PT1(1) = Vcl(1, vi): PT1(2) = Vcl(2, vi): PT1(3) = Vcl(3, vi)
        Call VBFAC(PT1(), Ctr, Vcl(), VM(), BF(), FC(), Front, Topnv)
        If (IERR <> 0) Then Exit Sub
'
        Call NWTHOU(I, NPT, Sizht, Nbf, Nfc, Maxbf, Maxfc, BF(), FC(), HT(), Ntetra, _
                    hdavbf, hdavfc, Front, Back, bfi, IERR)
        If (IERR <> 0) Then Exit Sub
'
        Call SWAPES(False, I, NPT, Sizht, Nfc, Maxfc, Vcl(), VM(), BF(), FC(), HT(), _
                    Ntetra, hdavfc, Front, Back, J, IERR)
        If (IERR <> 0) Then Exit Sub
    Next I
'
    Nface = Nfc
    Ptr = hdavfc
'
    Do While (Ptr <> 0)
        Nface = Nface - 1
        Ptr = -FC(1, Ptr)
    Loop
'
    FC(7, 1) = hdavbf
    FC(7, 2) = hdavfc
'
'600 format (/1x,'dtris3: first tetrahedron: ',4i7/4x,'i3, i4 =',2i7)
'610 format (/1x,'step',i7,':   vertex i =',i7)
'
'
'
End Sub
Private Sub DHPSRT(ByVal K&, ByVal N&, ByVal LDA&, a#(), ByRef MAP&())
'
'******************************************************************************
'
'   DHPSRT sorts a list of double precision points in KD.
'
'
'   Purpose:
'
'    Use heapsort to obtain the permutation of N K-dimensional
'    double precision points so that the points are in lexicographic
'    increasing order.
'
'   Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'   Parameters:
'
'    Input, K - dimension of points.
'
'    Input, N - number of points.
'
'    Input, LDA - leading dimension of array A in calling routine; should
'    be >= K.
'
'    Input, A(1:K,1:*) - array of >= N K-D double precision points.
'
'    Input/output, MAP(1:N).  On input, he points of A with indices
'    MAP(1), MAP(2), ..., MAP(N) are to be sorted.  On output, the elements
'    are permuted so that A(*,MAP(1)) <= A(*,MAP(2)) <= ... <= A(*,MAP(N)).
'
    Dim I&, T&
'
    For I = N \ 2 To 1 Step -1
        Call DSFTDW(I, N, K, LDA, a(), MAP())
    Next I
'
    For I = N To 2 Step -1
        T = MAP(1)
        MAP(1) = MAP(I)
        MAP(I) = T
        Call DSFTDW(1, I - 1, K, LDA, a(), MAP())
    Next I
'
'
'
End Sub
Private Sub DSFTDW(ByVal L, ByVal u, ByVal K, ByVal LDA, a#(), ByRef MAP&())
'
'******************************************************************************
'
'  DSFTDW does one step of the heap sort algorithm for double precision data.
'
'
'  Purpose:
'
'    Sift A(*,MAP(L)) down a heap of size U.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, L, U - lower and upper index of part of heap.
'
'    Input, K - dimension of points.
'
'    Input, LDA - leading dimension of array A in calling routine.
'
'    Input, A(1:K,1:*), see routine DHPSRT.
'
'    Input/output, MAP(1:*) - see routine DHPSRT.
'
'
  Dim I&, J&, T&
'
    I = L
    J = 2 * I
    T = MAP(I)
'
    Do
        If (J > u) Then
            Exit Do
        End If
'
        If (J < u) Then
            'If (DLESS(K, a(1, MAP(J)), a(1, MAP(J + 1)))) Then
            If (DLESS(K, a(), 0, MAP(J), a(), 0, MAP(J + 1))) Then
                J = J + 1
            End If
        End If
'
        'If (DLESS(K, a(1, MAP(J)), a(1, T))) Then
        If (DLESS(K, a(), 0, MAP(J), a(), 0, T)) Then
            Exit Do
        End If
'
        MAP(I) = MAP(J)
        I = J
        J = 2 * I
    Loop
'
    MAP(I) = T
'
'
'
End Sub
Private Function DLESS(ByVal K&, P#(), ByVal I1&, ByVal J1&, _
    Q#(), ByVal I2&, ByVal J2&) As Boolean
'
'******************************************************************************
'
'   DLESS determines the lexicographically lesser of two double precision values.
'
'
'   Purpose:
'
'    Determine whether P is lexicographically less than Q in
'    floating point arithmetic.
'
'   Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada    T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'   Parameters:
'
'    Input, K - dimension of points.
'
'    Input, P(1:K),Q(1:K) - two K-dimensional double precision points.
'
'    Output, DLESS - .TRUE. if P < Q, .FALSE. otherwise.
'
    Dim I&, cmax#, tol#
'
    tol = 100# * Epsilon '(Tol)
'
    For I = 1 To K
        cmax = DMAX1(Abs(P(I + I1, J1)), Abs(Q(I + I2, J2)))
'
        If (Abs(P(I + I1, J1) - Q(I + I2, J2)) <= tol * cmax Or cmax <= tol) Then
            GoTo 10 'cycle
        End If
'
        If (P(I + I1, J1) < Q(I + I2, J2)) Then
            DLESS = True
        Else
            DLESS = False
        End If
'
        Exit Function
10
    Next I
'
    DLESS = False
'
'
'
End Function
Private Sub SWAPES(ByVal bndcon As Boolean, ByVal I&, NPT&, ByVal Sizht&, _
    ByRef Nfc&, ByVal Maxfc&, Vcl#(), VM&(), ByRef BF&(), ByRef FC&(), _
    ByRef HT&(), ByRef Ntetra&, ByRef hdavfc&, ByRef Front&, ByRef Back&, _
    ByRef Ifac&, ByRef IERR&)
'
'******************************************************************************
'
'  SWAPES swaps faces in a 3D triangulation.
'
'
'  Purpose:
'
'    Swap faces (apply local transformations) in 3D triangulation
'    based on empty circumsphere criterion until (nearly)
'    all faces are locally optimal, where I is index of new vertex
'    added to triangulation or 0 if initial triangulation given.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, BNDCON - .TRUE. iff boundary faces are constrained (i.e. not
'    swapped by local transformations).
'
'    Input, I - local index of next vertex inserted in triangulation, or 0;
'    if positive, it is assumed I is largest index so far.
'
'    Input, NPT - number of 3D points to be triangulated.
'
'    Input, SIZHT - size of hash table HT.
'
'    Input/output, NFC - number of positions used in FC array.
'
'    Input, MAXFC - maximum size available for FC array.
'
'    Input, VCL(1:3,1:*) - vertex coordinate list.
'
'    Input, VM(1:NPT) - indices of vertices of VCL being triangulated.
'
'    Input/output, BF(1:3,1:*) -  array of boundary face records; see DTRIS3.
'
'    Input/output, FC(1:7,1:MAXFC) - array of face records; see routine DTRIS3.
'
'    Input/output, HT(0:SIZHT-1) - hash table using direct chaining.
'
'    Input/output, NTETRA - number of tetrahedra in triangulation.
'
'    Input/output, HDAVFC - head pointer to available FC records.
'
'    Input/output, FRONT, BACK - indices of front and back of queue of interior
'    faces for which sphere test is applied.
'
'    Output, IFAC - index of last face for which sphere test applied, or 0.
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
'
    Dim a&, aa&, b&, c&, d&, dd&, e&, F&, g&, In1&, Ind&, ind1&, ind2&
    Dim J&, K&, kneg&, kzero&, va&, vb&, vc&, Vd&, ve&
    Dim bfx&(1 To 2), Indx&(1 To 2), nbr&(1 To 2, 1 To 2)
    Dim Alpha#(1 To 4), Center#(1 To 3), Radsq#
    Dim Degen As Boolean
    Const msglvl& = 0
    Dim PT1#(1 To 3), PT2#(1 To 3), PT3#(1 To 3), PT4#(1 To 3), PT5#(1 To 3) ' Vettori d' appoggio.
'
    IERR = 0
    Ifac = 0
    
10  'continue
'
    If (Front = 0) Then Exit Sub
    Ind = Front
    Front = FC(7, Ind)
'
    If (FC(2, Ind) = 0) Then
        If (Ind = Nfc) Then
            Nfc = Nfc - 1
        Else
            FC(1, Ind) = -hdavfc
            hdavfc = Ind
        End If
'
        GoTo 10
    End If
'
    Ifac = Ind
    FC(7, Ind) = -1
    a = FC(1, Ind)
    b = FC(2, Ind)
    c = FC(3, Ind)
    d = FC(4, Ind)
    e = FC(5, Ind)
    va = VM(a)
    vb = VM(b)
    vc = VM(c)
    Vd = VM(d)
    ve = VM(e)
'
    If (msglvl = 4) Then
        'write ( *,600) Ind,a,b,c,d,e
    End If
'
    'Call CCSPH(True, Vcl(1, va), Vcl(1, vb), Vcl(1, vc), Vcl(1, Vd),Vcl(1, ve), _
                center, radsq, in1)
    PT1(1) = Vcl(1, va): PT1(2) = Vcl(2, va): PT1(3) = Vcl(3, va)
    PT2(1) = Vcl(1, vb): PT2(2) = Vcl(2, vb): PT2(3) = Vcl(3, vb)
    PT3(1) = Vcl(1, vc): PT3(2) = Vcl(2, vc): PT3(3) = Vcl(3, vc)
    PT4(1) = Vcl(1, Vd): PT4(2) = Vcl(2, Vd): PT4(3) = Vcl(3, Vd)
    PT5(1) = Vcl(1, ve): PT5(2) = Vcl(2, ve): PT5(3) = Vcl(3, ve)
    Call CCSPH(True, PT1(), PT2(), PT3(), PT4(), PT5(), Center(), Radsq, In1)
'
    If (In1 = 2) Then
        IERR = 301
        Exit Sub
    End If
'
    If (In1 >= 1) Then
        'Call BARYTH(Vcl(1, va), Vcl(1, vb), Vcl(1, vc), Vcl(1, Vd),Vcl(1, ve), _
                     alpha, degen)
        PT1(1) = Vcl(1, va): PT1(2) = Vcl(2, va): PT1(3) = Vcl(3, va)
        PT2(1) = Vcl(1, vb): PT2(2) = Vcl(2, vb): PT2(3) = Vcl(3, vb)
        PT3(1) = Vcl(1, vc): PT3(2) = Vcl(2, vc): PT3(3) = Vcl(3, vc)
        PT4(1) = Vcl(1, Vd): PT4(2) = Vcl(2, Vd): PT4(3) = Vcl(3, Vd)
        PT5(1) = Vcl(1, ve): PT5(2) = Vcl(2, ve): PT5(3) = Vcl(3, ve)
        Call BARYTH(PT1(), PT2(), PT3(), PT4(), PT5(), Alpha(), Degen)
'
        If (Degen) Then
            IERR = 301
            Exit Sub
        ElseIf (Alpha(4) > 0#) Then
            IERR = 309
            Exit Sub
        End If
'
        kneg = 1
        kzero = 0
'
        For J = 1 To 3
            If (Alpha(J) < 0#) Then
                kneg = kneg + 1
            ElseIf (Alpha(J) = 0#) Then
                kzero = kzero + 1
            End If
        Next J
'
        ' Swap 2 tetrahedra for 3.
'
        If (kneg = 1 And kzero = 0) Then
            Call UPDATF(a, b, d, c, e, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
            Call UPDATF(a, c, d, b, e, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
            Call UPDATF(b, c, d, a, e, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
            Call UPDATF(a, b, e, c, d, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
            Call UPDATF(a, c, e, b, d, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
            Call UPDATF(b, c, e, a, d, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
'
            If (IERR <> 0) Then Exit Sub
'
            Call HTDEL(Ind, NPT, Sizht, FC(), HT())
            Call HTINS(Ind, a, d, e, b, c, NPT, Sizht, FC(), HT())
            Call AVAILF(hdavfc, Nfc, Maxfc, FC(), Ind, IERR)
'
            If (IERR <> 0) Then Exit Sub
'
            Call HTINS(Ind, b, d, e, a, c, NPT, Sizht, FC(), HT())
            Call AVAILF(hdavfc, Nfc, Maxfc, FC(), Ind, IERR)
'
            If (IERR <> 0) Then Exit Sub
'
            Call HTINS(Ind, c, d, e, a, b, NPT, Sizht, FC(), HT())
            Ntetra = Ntetra + 1
'
            If (msglvl = 4) Then
                'write ( *,610)
            End If
'
            ' Swap 3 tetrahedra for 2 if possible. Relabel so edge
            ' AB would be deleted. Swap if ABDE is in1 current triangulation.
'
        ElseIf (kneg = 2 And kzero = 0) Then
            If (Alpha(1) < 0#) Then
                Call I_SWAP(a, c)
            ElseIf (Alpha(2) < 0#) Then
                Call I_SWAP(b, c)
            End If
'
            ind1 = HTSRC(a, b, d, NPT, Sizht, FC(), HT())
            If (ind1 <= 0) Then GoTo 50
'
            If (FC(4, ind1) = e Or FC(5, ind1) = e) Then
                Call UPDATF(a, c, d, b, e, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                Call UPDATF(a, c, e, b, d, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                Call UPDATF(a, d, e, b, c, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                Call UPDATF(b, c, d, a, e, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                Call UPDATF(b, c, e, a, d, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                Call UPDATF(b, d, e, a, c, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                If (IERR <> 0) Then Exit Sub
                Call HTDEL(Ind, NPT, Sizht, FC(), HT())
                Call HTINS(Ind, c, d, e, a, b, NPT, Sizht, FC(), HT())
                Call HTDEL(ind1, NPT, Sizht, FC(), HT())
'
                If (FC(7, ind1) >= 0) Then
                    FC(2, ind1) = 0
                Else
                    If (ind1 = Nfc) Then
                        Nfc = Nfc - 1
                    Else
                        FC(1, ind1) = -hdavfc
                        hdavfc = ind1
                    End If
                End If
'
                ind1 = HTSRC(a, b, e, NPT, Sizht, FC(), HT())
                If (ind1 <= 0) Then GoTo 50
                Call HTDEL(ind1, NPT, Sizht, FC(), HT())
'
                If (FC(7, ind1) >= 0) Then
                    FC(2, ind1) = 0
'
                Else
                    If (ind1 = Nfc) Then
                        Nfc = Nfc - 1
                    Else
                        FC(1, ind1) = -hdavfc
                        hdavfc = ind1
                    End If
                End If
'
                Ntetra = Ntetra - 1
'
                If (msglvl = 4) Then
                    'write ( *,620) c,d,e
                End If
'
            Else
                If (msglvl = 4) Then
                    'write ( *,630) a,b,d,e
                End If
            End If
'
            ' Coplanar faces: swap 2 tetrahedra for 2 if boundary faces
            ' (and BNDCON is .FALSE.), else do pair of 2 for 2 swaps if
            ' possible.  Relabel vertices so that DE intersects AB.
            ' Also swap if necessary to make A < B and D < E.
'
        ElseIf (kneg = 1 And kzero = 1) Then
            If (Alpha(1) = 0#) Then
                Call I_SWAP(a, c)
'
            ElseIf (Alpha(2) = 0#) Then
                Call I_SWAP(b, c)
            End If
'
            If (a > b) Then
                Call I_SWAP(a, b)
            End If
'
            If (d > e) Then
                Call I_SWAP(d, e)
            End If
'
            ind1 = HTSRC(a, b, d, NPT, Sizht, FC(), HT())
            ind2 = HTSRC(a, b, e, NPT, Sizht, FC(), HT())
            If (ind1 <= 0 Or ind2 <= 0) Then GoTo 50
'
            If (FC(4, ind1) = c) Then
                F = FC(5, ind1)
            Else
                F = FC(4, ind1)
            End If
'
            If (FC(4, ind2) = c) Then
                g = FC(5, ind2)
            Else
                g = FC(4, ind2)
            End If
'
            If (F <= 0 And g <= 0) Then
                If (Not bndcon) Then
                    Call UPDATF(a, c, d, b, e, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                    Call UPDATF(a, c, e, b, d, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                    Call UPDATF(b, c, d, a, e, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                    Call UPDATF(b, c, e, a, d, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
'
                    If (IERR <> 0) Then Exit Sub
'
                    Call HTDEL(Ind, NPT, Sizht, FC(), HT())
                    Call HTINS(Ind, c, d, e, a, b, NPT, Sizht, FC(), HT())
                    Call HTDEL(ind1, NPT, Sizht, FC(), HT())
                    Call HTINS(ind1, a, d, e, c, FC(5, ind1), NPT, Sizht, FC(), HT())
                    Call HTDEL(ind2, NPT, Sizht, FC(), HT())
                    Call HTINS(ind2, b, d, e, c, FC(5, ind2), NPT, Sizht, FC(), HT())
                    Indx(1) = ind1
                    Indx(2) = ind2
                    bfx(1) = -FC(5, ind1)
                    bfx(2) = -FC(5, ind2)
                    dd = d
'
                    For J = 1 To 2
                        If (J = 2) Then dd = e
'
                        If (dd < a) Then
                            nbr(J, 1) = BF(3, bfx(J))
                            nbr(J, 2) = BF(2, bfx(J))
                        ElseIf (dd < b) Then
                            nbr(J, 1) = BF(3, bfx(J))
                            nbr(J, 2) = BF(1, bfx(J))
                        Else
                            nbr(J, 1) = BF(2, bfx(J))
                            nbr(J, 2) = BF(1, bfx(J))
                        End If
                    Next J
'
                    aa = a
                    K = -FC(5, nbr(1, 2))
'
                    For J = 1 To 2
                        If (J = 2) Then
                            aa = b
                            K = -FC(5, nbr(2, 1))
                        End If
'
                        If (aa < d) Then
                            BF(1, bfx(J)) = Indx(3 - J)
                            BF(2, bfx(J)) = nbr(2, J)
                            BF(3, bfx(J)) = nbr(1, J)
                        ElseIf (aa < e) Then
                            BF(1, bfx(J)) = nbr(2, J)
                            BF(2, bfx(J)) = Indx(3 - J)
                            BF(3, bfx(J)) = nbr(1, J)
                        Else
                            BF(1, bfx(J)) = nbr(2, J)
                            BF(2, bfx(J)) = nbr(1, J)
                            BF(3, bfx(J)) = Indx(3 - J)
                        End If
'
                        If (BF(1, K) = Indx(J)) Then
                            BF(1, K) = Indx(3 - J)
                        ElseIf (BF(2, K) = Indx(J)) Then
                            BF(2, K) = Indx(3 - J)
                        Else
                            BF(3, K) = Indx(3 - J)
                        End If
                    Next J
'
                    If (msglvl = 4) Then
                        'write ( *,640) a,b,d,e
                    End If
                End If
'
            ElseIf (F = g) Then
                Call UPDATF(a, c, d, b, e, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                Call UPDATF(a, c, e, b, d, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                Call UPDATF(b, c, d, a, e, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                Call UPDATF(b, c, e, a, d, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                Call UPDATF(a, d, F, b, e, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                Call UPDATF(a, e, F, b, d, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                Call UPDATF(b, d, F, a, e, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
                Call UPDATF(b, e, F, a, d, I, NPT, Sizht, Front, Back, FC(), HT(), IERR)
'
                If (IERR <> 0) Then Exit Sub
'
                Call HTDEL(Ind, NPT, Sizht, FC(), HT())
                Call HTINS(Ind, c, d, e, a, b, NPT, Sizht, FC(), HT())
                Ind = HTSRC(a, b, F, NPT, Sizht, FC(), HT())
                If (Ind <= 0) Then GoTo 50
                Call HTDEL(Ind, NPT, Sizht, FC(), HT())
'
                If (FC(7, Ind) >= 0) Then
                    FC(2, Ind) = 0
                    Call AVAILF(hdavfc, Nfc, Maxfc, FC(), Ind, IERR)
                    If (IERR <> 0) Then Exit Sub
                End If
'
                Call HTINS(Ind, d, e, F, a, b, NPT, Sizht, FC(), HT())
                Call HTDEL(ind1, NPT, Sizht, FC(), HT())
                J = FC(7, ind1)
                Call HTINS(ind1, a, d, e, c, F, NPT, Sizht, FC(), HT())
                FC(7, ind1) = J
                Call HTDEL(ind2, NPT, Sizht, FC(), HT())
                J = FC(7, ind2)
                Call HTINS(ind2, b, d, e, c, F, NPT, Sizht, FC(), HT())
                FC(7, ind2) = J
'
                If (I <= 0 And FC(7, ind1) = -1) Then
                    FC(7, ind1) = 0
                    If (Front = 0) Then
                        Front = ind1
                    Else
                        FC(7, Back) = ind1
                    End If
                    Back = ind1
                End If
'
                If (I <= 0 And FC(7, ind2) = -1) Then
                    FC(7, ind2) = 0
                    If (Front = 0) Then
                        Front = ind2
                    Else
                        FC(7, Back) = ind2
                    End If
                    Back = ind2
                End If
'
                If (msglvl = 4) Then
                    'write ( *,650) a,b,d,e,f
                End If
'
            Else
                If (msglvl = 4) Then
                    'write ( *,660) a,b,d,e,f,g
                End If
            End If
        End If
    End If
'
    GoTo 10
'
50  'continue
'
    IERR = 300
'
'600 format (1x,'index =',i7,' : ',5i7)
'610 format (4x,'swap 2-3')
'620 format (4x,'swap 3-2 with new common face:',3i7)
'630 format (4x,'swap 3-2 not poss, tetra missing:',4i7)
'640 format (4x,'swap 2-2: edge ',2i7,' repl by ',2i7)
'650 format (4x,'swap 4-4: edge ',2i7,' repl by ',2i7,'   f =',i7)
'660 format (4x,'swap 4-4 not poss: a,b,d,e,f,g =',6i7)
'
'
'
End Sub
Private Sub AVAILF(ByRef hdavfc&, ByRef Nfc&, ByVal Maxfc&, FC&(), _
    ByRef Ind&, ByRef IERR&)
'
'******************************************************************************
'
'  AVAILF returns the index of the next available record in the FC array.
'
'
'  Purpose:
'
'    Return index of next available record in FC array,
'    either HDAVFC or NFC+1.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input/output, HDAVFC - head pointer of available records in FC.
'
'    Input/output, NFC - current number of records used in FC.
'
'    Input, integer MAXFC, the maximum number of records available in FC.
'
'    Input, FC(1:7,1:*) - array of face records; see routine DTRIS3.
'
'    Output, integer IND, the index of available record (if FC not full).
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
'
    IERR = 0
'
    If (hdavfc <> 0) Then
        Ind = hdavfc
        hdavfc = -FC(1, hdavfc)
    Else
        If (Nfc >= Maxfc) Then
            IERR = 11
        Else
            Nfc = Nfc + 1
            Ind = Nfc
        End If
    End If
'
'
'
End Sub
Private Sub HTDEL(ByVal Ind&, ByVal N&, ByVal P&, ByRef FC&(), ByRef HT&())
'
'******************************************************************************
'
'  HTDEL deletes a record from the hash table.
'
'
'  Purpose:
'
'    Delete record FC(1:7,IND) from hash table HT.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, IND - index of FC array.
'
'    Input, N - upper bound on vertex indices.
'
'    Input, P - size of hash table.
'
'    Input/output, FC(1:7,1:*) - array of face records; see routine DTRIS3.
'    On output, one link in FC is updated.
'
'    Input/output, HT(0:P-1) - hash table using direct chaining.  On output,
'    one link in HT is updated.
'
'
    Dim K&, Ptr&
'
    K = (FC(1, Ind) * N + FC(2, Ind)) Mod P
    K = (K * N + FC(3, Ind)) Mod P
    Ptr = HT(K)
'
    If (Ptr = Ind) Then
        HT(K) = FC(6, Ind)
'
    Else
        Do While (FC(6, Ptr) <> Ind)
            Ptr = FC(6, Ptr)
        Loop
        FC(6, Ptr) = FC(6, Ind)
    End If
'
'
'
End Sub
Private Sub UPDATF(ByVal a&, ByVal b&, ByVal c&, ByVal d&, ByVal e&, _
    ByVal I&, ByVal N&, ByVal P&, ByRef Front&, ByRef Back&, _
    FC&(), HT&(), ByRef IERR)
'
'******************************************************************************
'
'  UPDATF updates a record in FC after a local transformation.
'
'
'  Purpose:
'
'    Update record in FC due to a local transformation.
'    Tetrahedron ABCD becomes ABCE. Add face ABC to queue if it is
'    interior face, not yet in queue, and its largest index isn't I.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, A, B, C - first 3 fields of FC record (in any order).
'
'    Input, D, E - fourth vertex indices of old and new tetrahedrons.
'
'    Input, I - vertex index determining whether face put on queue.
'
'    Input, N - upper bound on vertex indices.
'
'    Input, P - size of hash table.
'
'    Input/output, FRONT, BACK - front and back pointers of queue.
'
'    Input, FC(1:7,1:*) - array of face records; see routine DTRIS3.
'
'    Input, HT(0:P-1) - hash table using direct chaining.
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
'
    Dim Ind&
'
    IERR = 0
    Ind = HTSRC(a, b, c, N, P, FC(), HT())
'
    If (Ind <= 0) Then
        IERR = 300
        Exit Sub
    End If
'
    If (FC(4, Ind) = d) Then
        FC(4, Ind) = e
    Else
        FC(5, Ind) = e
    End If
'
    If (FC(7, Ind) = -1 And FC(3, Ind) <> I And FC(5, Ind) > 0) Then
        FC(7, Ind) = 0
'
        If (Front = 0) Then
            Front = Ind
        Else
            FC(7, Back) = Ind
        End If
'
        Back = Ind
    End If
'
'
'
End Sub
Private Sub BARYTH(a#(), b#(), c#(), d#(), e#(), ByRef Alpha#(), ByRef Degen As Boolean)
'
'******************************************************************************
'
'  BARYTH computes barycentric coordinates of a point in 3D.
'
'
'  Purpose:
'
'    Compute barycentric coordinates of 3D point with respect
'    to four vertices of a tetrahedron.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, A(1:3),B(1:3),C(1:3),D(1:3) - 4 vertices of tetrahedron.
'
'    Input, E(1:3) - fifth point for which barycentric coordinates found
'
'    Output, ALPHA(1:4) - scaled barycentric coordinates (if DEGEN = .FALSE.)
'    such that E = (ALPHA(1)*A + ALPHA(2)*B + ALPHA(3)*C +
'    ALPHA(4)*D)/DET where DET = 6 * (volume of tetra ABCD);
'    an ALPHA(I) may be set to 0 after tolerance test to
'    indicate that E is coplanar with a face, so sum of
'    ALPHA(I)/DET may not be 1; if the actual barycentric
'    coordinates rather than just their signs are needed,
'    modify this routine to divide ALPHA(I) by DET.
'
'    Output, DEGEN - .TRUE. iff A,B,C,D are coplanar.
'
'
    Dim I&
    Dim amax#, bmax#, cmax#, cp1#, cp2#, cp3#, det#, dmax#, emax#, tol#
    Dim da#(1 To 3), db#(1 To 3), dc#(1 To 3), de#(1 To 3)
    Dim ea#(1 To 3), eb#(1 To 3), ec#(1 To 3)
'
    tol = 100# * Epsilon
    Degen = False
'
    'da(1:3) = a(1:3) - d(1:3)
    'db(1:3) = b(1:3) - d(1:3)
    'dc(1:3) = c(1:3) - d(1:3)
    For I = 1 To 3
        da(I) = a(I) - d(I)
        db(I) = b(I) - d(I)
        dc(I) = c(I) - d(I)
    Next I
'
    amax = DMAX1(Abs(a(1)), Abs(a(2)), Abs(a(3)))
    bmax = DMAX1(Abs(b(1)), Abs(b(2)), Abs(b(3)))
    cmax = DMAX1(Abs(c(1)), Abs(c(2)), Abs(c(3)))
    dmax = DMAX1(Abs(d(1)), Abs(d(2)), Abs(d(3)))
    cp1 = db(2) * dc(3) - db(3) * dc(2)
    cp2 = db(3) * dc(1) - db(1) * dc(3)
    cp3 = db(1) * dc(2) - db(2) * dc(1)
    det = da(1) * cp1 + da(2) * cp2 + da(3) * cp3
'
    If (Abs(det) <= 0.01 * tol * DMAX1(amax, bmax, cmax, dmax)) Then
        Degen = True
        Exit Sub
    End If
'
    'de(1:3) = e(1:3) - d(1:3)
    'ea(1:3) = a(1:3) - e(1:3)
    'eb(1:3) = b(1:3) - e(1:3)
    'ec(1:3) = c(1:3) - e(1:3)
    For I = 1 To 3
        de(I) = e(I) - d(I)
        ea(I) = a(I) - e(I)
        eb(I) = b(I) - e(I)
        ec(I) = c(I) - e(I)
    Next I
'
    Alpha(1) = de(1) * cp1 + de(2) * cp2 + de(3) * cp3
    cp1 = da(2) * de(3) - da(3) * de(2)
    cp2 = da(3) * de(1) - da(1) * de(3)
    cp3 = da(1) * de(2) - da(2) * de(1)
    Alpha(2) = dc(1) * cp1 + dc(2) * cp2 + dc(3) * cp3
    Alpha(3) = db(1) * cp1 + db(2) * cp2 + db(3) * cp3
    Alpha(4) = ea(1) * (eb(2) * ec(3) - eb(3) * ec(2)) + ea(2) * (eb(3) * ec(1) _
             - eb(1) * ec(3)) + ea(3) * (eb(1) * ec(2) - eb(2) * ec(1))
'
    If (det < 0#) Then
        Alpha(1) = -Alpha(1)
        Alpha(2) = -Alpha(2)
        Alpha(4) = -Alpha(4)
    Else
        Alpha(3) = -Alpha(3)
    End If
'
    emax = DMAX1(Abs(e(1)), Abs(e(2)), Abs(e(3)))
'
    If (Abs(Alpha(1)) <= tol * DMAX1(bmax, cmax, dmax, emax)) Then
        Alpha(1) = 0#
    End If
'
    If (Abs(Alpha(2)) <= tol * DMAX1(amax, cmax, dmax, emax)) Then
        Alpha(2) = 0#
    End If
'
    If (Abs(Alpha(3)) <= tol * DMAX1(amax, bmax, dmax, emax)) Then
        Alpha(3) = 0#
    End If
'
    If (Abs(Alpha(4)) <= tol * DMAX1(amax, bmax, cmax, emax)) Then
        Alpha(4) = 0#
    End If
'
'
'
End Sub
Private Sub CCSPH(ByVal Intest As Boolean, a#(), b#(), c#(), d#(), e#(), _
    ByRef Center#(), ByRef Radsq#, ByRef In1&)
'
'******************************************************************************
'
'  CCSPH finds the circumsphere through the vertices of a tetrahedron.
'
'
'  Purpose:
'
'    Find center and square of radius of circumsphere through
'    four vertices of a tetrahedron, and possibly determine whether
'    a fifth 3D point is inside sphere.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, INTEST - .TRUE. iff test for fifth point in sphere to be made.
'
'    Input, A(1:3), B(1:3), C(1:3), D(1:3) - 4 vertices of tetrahedron.
'
'    Input, E(1:3) - fifth point; referenced iff INTEST is .TRUE.
'
'    Output, CENTER(1:3) - center of sphere; undefined if A,B,C,D coplanar.
'
'    Output, RADSQ - square of radius of sphere; -1 if A,B,C,D coplanar.
'
'    Output, IN - contains following value if INTEST is .TRUE.:
'    2 if A,B,C,D coplanar; 1 if E inside sphere;
'    0 if E on sphere; -1 if E outside sphere
'
'
    Dim I&
    Dim cmax#, cp1#, cp2#, cp3#, det#, dsq#, tol#
    Dim da#(1 To 3), db#(1 To 3), dc#(1 To 3), rhs#(1 To 3)
'
    tol = 100# * Epsilon
'
    'da(1:3) = a(1:3) - d(1:3)
    'db(1:3) = b(1:3) - d(1:3)
    'dc(1:3) = c(1:3) - d(1:3)
    For I = 1 To 3
        da(I) = a(I) - d(I)
        db(I) = b(I) - d(I)
        dc(I) = c(I) - d(I)
    Next I
'
    rhs(1) = 0.5 * (da(1) ^ 2 + da(2) ^ 2 + da(3) ^ 2)
    rhs(2) = 0.5 * (db(1) ^ 2 + db(2) ^ 2 + db(3) ^ 2)
    rhs(3) = 0.5 * (dc(1) ^ 2 + dc(2) ^ 2 + dc(3) ^ 2)
'
    cmax = DMAX1( _
    Abs(a(1)), Abs(a(2)), Abs(a(3)), _
    Abs(b(1)), Abs(b(2)), Abs(b(3)), _
    Abs(c(1)), Abs(c(2)), Abs(c(3)), _
    Abs(d(1)), Abs(d(2)), Abs(d(3)))
'
    cp1 = db(2) * dc(3) - dc(2) * db(3)
    cp2 = dc(2) * da(3) - da(2) * dc(3)
    cp3 = da(2) * db(3) - db(2) * da(3)
    det = da(1) * cp1 + db(1) * cp2 + dc(1) * cp3
'
    If (Abs(det) <= 0.01 * tol * cmax) Then
        Radsq = -1#
        In1 = 2
        Exit Sub
    End If
'
    Center(1) = (rhs(1) * cp1 + rhs(2) * cp2 + rhs(3) * cp3) / det
    cp1 = db(1) * rhs(3) - dc(1) * rhs(2)
    cp2 = dc(1) * rhs(1) - da(1) * rhs(3)
    cp3 = da(1) * rhs(2) - db(1) * rhs(1)
    Center(2) = (da(3) * cp1 + db(3) * cp2 + dc(3) * cp3) / det
    Center(3) = -(da(2) * cp1 + db(2) * cp2 + dc(2) * cp3) / det
    Radsq = Center(1) ^ 2 + Center(2) ^ 2 + Center(3) ^ 2
'
    For I = 1 To 3
        Center(I) = Center(I) + d(I)
    Next I
'
    If (Intest) Then
        'dsq = sum ( ( e(1:3) - center(1:3) )^2 )
        dsq = 0#
        For I = 1 To 3
            dsq = dsq + (e(I) - Center(I)) ^ 2
        Next I
'
        If (dsq > (1# + tol) * Radsq) Then
            In1 = -1
        ElseIf (dsq < (1# - tol) * Radsq) Then
            In1 = 1
        Else
            In1 = 0
        End If
    End If
'
'
'
End Sub
Private Sub NWTHED(ByVal I&, ByVal Ifac&, ByVal IEDG&, ByVal NPT&, ByVal Sizht&, _
    ByRef Nbf&, ByRef Nfc&, ByVal Maxbf&, ByVal Maxfc&, ByRef BF&(), ByRef FC&(), _
    ByRef HT&(), ByRef Ntetra&, ByRef hdavbf&, ByRef hdavfc&, _
    ByRef Front&, ByRef Back&, ByRef IERR&)
'
'******************************************************************************
'
'  NWTHED creates new tetrahedra from the insertion of a vertex, in 3D.
'
'
'  Purpose:
'
'    Create new tetrahedra in 3D triangulation from the
'    insertion of vertex I on edge FC(IEDG,IFAC).
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, I - (local) index of next vertex inserted in triangulation;
'    it is assumed I is largest index so far.
'
'    Input, IFAC, IEDG - edge containing I is FC(IEDG,IFAC); 1 <= IEDG <= 3.
'
'    Input, NPT - number of 3D points to be triangulated.
'
'    Input, SIZHT - size of hash table HT.
'
'    Input/output, NBF - number of positions used in BF array.
'
'    Input/output, NFC - number of positions used in FC array.
'
'    Input, MAXBF - maximum size available for BF array.
'
'    Input, MAXFC - maximum size available for FC array.
'
'    Input/output, BF(1:3,1:MAXBF) -  array of boundary face records;
'    see DTRIS3.
'
'    Input/output, FC(1:7,1:MAXFC) - array of face records; see routine DTRIS3.
'
'    Input/output, HT(0:SIZHT-1) - hash table using direct chaining.
'
'    Input/output, NTETRA - number of tetrahedra in triangulation.
'
'    Input/output, HDAVBF - head pointer to available BF records.
'
'    Input/output, HDAVFC - head pointer to available FC records.
'
'    Output, FRONT, BACK - indices of front and back of queue of interior
'    faces ABC such that ABCI is a new tetrahedron.
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
    Dim a&, aa&, b&, Bb&, c&, cp&, csav&, d&, e&, Ind&, inew&
    Dim bfn&(1 To 2), bfo&(1 To 2), fcn&(1 To 2), fco&(1 To 2)
    Dim J&, K&, L&, nbrac&, nbrbc&, nfcin&
    Dim bedge As Boolean
    Const msglvl& = 0
'
    IERR = 0
    Front = 0
    Back = 0
'
    If (IEDG = 1) Then
        a = FC(1, Ifac)
        b = FC(2, Ifac)
        c = FC(3, Ifac)
    ElseIf (IEDG = 2) Then
        a = FC(2, Ifac)
        b = FC(3, Ifac)
        c = FC(1, Ifac)
    Else
        a = FC(1, Ifac)
        b = FC(3, Ifac)
        c = FC(2, Ifac)
    End If
'
    csav = c
    d = FC(4, Ifac)
'
    ' Determine faces incident on edge AB in circular order, and store
    ' their indices at end of FC array.
'
    bedge = False
    Ind = Ifac
    K = 0
'
    Do
        K = K + 1
'
        If (Nfc + K > Maxfc) Then
            IERR = 11
            Exit Sub
        End If
'
        FC(1, Nfc + K) = Ind
        If (bedge) Then GoTo 20
        If (d = csav) Then GoTo 50
        Ind = HTSRC(a, b, d, NPT, Sizht, FC(), HT())
'
        If (Ind <= 0) Then
            IERR = 300
            Exit Sub
        End If
'
        If (FC(5, Ind) <= 0) Then
            bedge = True
            cp = d
        ElseIf (FC(4, Ind) = c) Then
            c = d
            d = FC(5, Ind)
        Else
        c = d
        d = FC(4, Ind)
        End If
    Loop
'
    ' Determine further faces in case of AB being a boundary edge.
'
20  'continue
'
    If (FC(5, Ifac) <= 0) Then GoTo 50
'
    L = K
'
    For J = 1 To K / 2
        e = FC(1, Nfc + J)
        FC(1, Nfc + J) = FC(1, Nfc + L)
        FC(1, Nfc + L) = e
        L = L - 1
    Next J
'
    c = csav
    csav = cp
    d = FC(5, Ifac)
'
    Do
        Ind = HTSRC(a, b, d, NPT, Sizht, FC(), HT())
'
        If (Ind <= 0) Then
            IERR = 300
            Exit Sub
        End If
'
        K = K + 1
'
        If (Nfc + K > Maxfc) Then
            IERR = 11
            Exit Sub
        End If
'
        FC(1, Nfc + K) = Ind
'
        If (FC(5, Ind) <= 0) Then
            Exit Do
        ElseIf (FC(4, Ind) = c) Then
            c = d
            d = FC(5, Ind)
        Else
            c = d
            d = FC(4, Ind)
            End If
    Loop
'
    ' Create new faces and tetrahedra, and add faces to queue.
'
50  'continue
    
    nfcin = Nfc
    Nfc = Nfc + K
    Ntetra = Ntetra + K
'
    If (bedge) Then
    Ntetra = Ntetra - 1
    fcn(1) = nfcin + 1
    fcn(2) = nfcin + K
    fco(1) = FC(1, nfcin + 1)
    fco(2) = FC(1, nfcin + K)
    bfo(1) = -FC(5, fco(1))
    bfo(2) = -FC(5, fco(2))
    End If
'
    For J = 1 To K
        inew = nfcin + J
        Ind = FC(1, inew)
'
        If (FC(1, Ind) = a) Then
            If (FC(2, Ind) = b) Then
                c = FC(3, Ind)
            Else
                c = FC(2, Ind)
            End If
        Else
            c = FC(1, Ind)
        End If
'
        d = FC(4, Ind)
        e = FC(5, Ind)
        Call HTDEL(Ind, NPT, Sizht, FC(), HT())
        Call HTINS(Ind, a, c, I, d, e, NPT, Sizht, FC(), HT())
        Call HTINS(inew, b, c, I, d, e, NPT, Sizht, FC(), HT())
'
        If (J = K) Then
            If (bedge) Then
                GoTo Cycle 'cycle
            End If
            d = csav
        ElseIf (J > 1) Then
            If (d = cp) Then d = e
        End If
'
        Call AVAILF(hdavfc, Nfc, Maxfc, FC(), Ind, IERR)
        If (IERR <> 0) Then Exit Sub
        Call HTINS(Ind, c, d, I, a, b, NPT, Sizht, FC(), HT())
        aa = a
        Bb = b
'
        For L = 1 To 2
            If (L = 2) Then
                aa = b
                Bb = a
            End If
'
            Ind = HTSRC(aa, c, d, NPT, Sizht, FC(), HT())
'
            If (Ind <= 0) Then
                IERR = 300
                Exit Sub
            End If
'
            If (FC(5, Ind) <= 0) Then
                FC(4, Ind) = I
            Else
                If (FC(4, Ind) = Bb) Then
                    FC(4, Ind) = I
                Else
                    FC(5, Ind) = I
                End If
                If (Front = 0) Then
                    Front = Ind
                Else
                    FC(7, Back) = Ind
                End If
                Back = Ind
            End If
'
            If (msglvl = 4) Then
                'write ( *,600) aa,c,d,i
            End If
        Next L
'
        cp = c
Cycle:
    Next J
'
    If (Front <> 0) Then
        FC(7, Back) = 0
    End If
'
    If (bedge) Then
        d = c
        c = csav
'
        If (hdavbf <> 0) Then
            bfn(1) = hdavbf
            hdavbf = -BF(1, hdavbf)
'
            If (hdavbf <> 0) Then
                bfn(2) = hdavbf
                hdavbf = -BF(1, hdavbf)
            Else
                Nbf = Nbf + 1
                bfn(2) = Nbf
            End If
'
        Else
            Nbf = Nbf + 2
            bfn(1) = Nbf - 1
            bfn(2) = Nbf
        End If
'
        If (Nbf > Maxbf) Then
            Nbf = Maxbf
            IERR = 12
            Exit Sub
        End If
'
        FC(5, nfcin + 1) = -bfn(1)
        FC(5, nfcin + K) = -bfn(2)
'
        For J = 1 To 2
            If (J = 2) Then c = d
'
            If (c < a) Then
                nbrac = BF(3, bfo(J))
                nbrbc = BF(2, bfo(J))
                BF(1, bfo(J)) = fco(3 - J)
                BF(2, bfo(J)) = fcn(J)
                BF(1, bfn(J)) = fcn(3 - J)
                BF(2, bfn(J)) = fco(J)
            ElseIf (c < b) Then
                nbrac = BF(3, bfo(J))
                nbrbc = BF(1, bfo(J))
                BF(1, bfo(J)) = fcn(J)
                BF(2, bfo(J)) = fco(3 - J)
                BF(1, bfn(J)) = fcn(3 - J)
                BF(2, bfn(J)) = fco(J)
            Else
                nbrac = BF(2, bfo(J))
                nbrbc = BF(1, bfo(J))
                BF(1, bfo(J)) = fcn(J)
                BF(2, bfo(J)) = fco(3 - J)
                BF(1, bfn(J)) = fco(J)
                BF(2, bfn(J)) = fcn(3 - J)
            End If
'
            BF(3, bfo(J)) = nbrac
            BF(3, bfn(J)) = nbrbc
            L = -FC(5, nbrbc)
'
            If (BF(1, L) = fco(J)) Then
                BF(1, L) = fcn(J)
            ElseIf (BF(2, L) = fco(J)) Then
                BF(2, L) = fcn(J)
            Else
                BF(3, L) = fcn(J)
            End If
        Next J
    End If
'
'600 format (1x,'new tetra: ',4i7)
'
'
'
End Sub
Private Sub I_SWAP(ByRef I&, ByRef J&)
'
'*******************************************************************************
'
'  I_SWAP swaps two integer values.
'
'
'  Modified:
'
'    30 November 1998
'
'  Author:
'
'    John Burkardt
'
'  Parameters:
'
'    Input/output, integer I, J.  On output, the values of I and
'    J have been interchanged.
'
'
    Dim K&
'
    K = I
    I = J
    J = K
'
'
'
End Sub
Private Sub NWTHOU(ByVal I&, ByVal NPT&, ByVal Sizht&, ByRef Nbf&, ByRef Nfc&, _
    ByVal Maxbf&, ByVal Maxfc&, ByRef BF&(), ByRef FC&(), ByRef HT&(), _
    ByRef Ntetra&, ByRef hdavbf&, ByRef hdavfc&, ByVal Front&, ByRef Back&, _
    ByRef bfi&, ByRef IERR&)
'
'******************************************************************************
'
'  NWTHOU creates new tetrahedra outside the current convex hull.
'
'
'  Purpose:
'
'    Create new tetrahedra in 3D triangulation outside
'    convex hull by joining vertex I to visible boundary faces.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, I - (local) index of next vertex inserted in triangulation.
'
'    Input, NPT - number of 3D points to be triangulated.
'
'    Input, SIZHT - size of hash table HT.
'
'    Input/output, NBF - number of positions used in BF array.
'
'    Input/output, NFC - number of positions used in FC array.
'
'    Input, MAXBF - maximum size available for BF array.
'
'    Input, MAXFC - maximum size available for FC array.
'
'    Input/output, BF(1:3,1:MAXBF) -  array of boundary face records;
'    see DTRIS3.
'
'    Input/output, FC(1:7,1:MAXFC) - array of face records; see routine DTRIS3.
'
'    Input/output, HT(0:SIZHT-1) - hash table using direct chaining.
'
'    Input/output, NTETRA - number of tetrahedra in triangulation.
'
'    Input/output, HDAVBF - head pointer to available BF records.
'
'    Input/output, HDAVFC - head pointer to available FC records.
'
'    Input, FRONT - index of front of queue (or top of stack) of visible
'    boundary faces.
'
'    Output, BACK - index of back of queue (or bottom of stack) of visible
'    boundary faces (which become interior faces).
'
'    Output, BFI - index of FC of a boundary face containing vertex I.
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
'
    Dim a&, b&, c&, d&, e&, Ind&, J&, K&, L&, bfnew&, nbr&, Ptr&
    Const msglvl& = 0
'
    ' For ABC in queue, form tetrahedron ABCI + add faces ABI, ACI, BCI.
    ' PTR, NBR, IND are indices of FC; K, L, BFNEW indices of BF.
'
    IERR = 0
    bfi = 0
    Ptr = Front
    
10  'continue
    
    Back = Ptr
    a = FC(1, Ptr)
    b = FC(2, Ptr)
    c = FC(3, Ptr)
    K = -FC(5, Ptr)
    FC(5, Ptr) = I
    Ntetra = Ntetra + 1
'
    If (msglvl = 4) Then
        'write ( *,600) a,b,c,i
    End If
'
    For e = 1 To 3
        If (e = 2) Then
            Call I_SWAP(a, b)
        ElseIf (e = 3) Then
            Call I_SWAP(a, c)
        End If
'
        nbr = BF(e, K)
'
         If (FC(7, nbr) <> -1) Then
            If (FC(5, nbr) = I) Then
                GoTo Cycle 'cycle
            End If
        End If
'
        Call AVAILF(hdavfc, Nfc, Maxfc, FC(), Ind, IERR)
        If (IERR <> 0) Then Exit Sub
        L = -FC(5, nbr)
'
        If (BF(1, L) = Ptr) Then
            J = 1
        ElseIf (BF(2, L) = Ptr) Then
            J = 2
        Else
            J = 3
        End If
'
        If (FC(7, nbr) <> -1) Then
            Call HTINS(Ind, b, c, I, a, FC(J, nbr), NPT, Sizht, FC(), HT())
'
        Else
            If (hdavbf <> 0) Then
                bfnew = hdavbf
                hdavbf = -BF(1, hdavbf)
            Else
                If (Nbf >= Maxbf) Then
                    IERR = 12
                    Exit Sub
                End If
                Nbf = Nbf + 1
                bfnew = Nbf
            End If
'
            If (bfi = 0) Then bfi = Ind
            Call HTINS(Ind, b, c, I, a, -bfnew, NPT, Sizht, FC(), HT())
            BF(J, L) = Ind
            BF(3, bfnew) = nbr
        End If
Cycle:
    Next e
'
    If (K = Nbf) Then
        Nbf = Nbf - 1
    Else
        BF(1, K) = -hdavbf
        hdavbf = K
    End If
'
    Ptr = FC(7, Ptr)
    If (Ptr <> 0) Then GoTo 10
'
    ' Set BF(1:2,BFNEW) fields for new boundary faces.
'
    Ptr = bfi
    a = FC(1, Ptr)
    J = 2
    
30  'continue
'
    b = FC(J, Ptr)
    c = FC(4, Ptr)
    
40  'continue
'
    nbr = HTSRC(a, c, I, NPT, Sizht, FC(), HT())
'
    If (nbr <= 0) Then
        IERR = 300
        Exit Sub
    End If
'
    If (FC(5, nbr) > 0) Then
        If (FC(4, nbr) = b) Then
            d = FC(5, nbr)
        Else
            d = FC(4, nbr)
        End If
        b = c
        c = d
        GoTo 40
    End If
'
    K = -FC(5, Ptr)
    L = -FC(5, nbr)
'
    If (FC(1, Ptr) = a) Then
        BF(2, K) = nbr
    Else
        BF(1, K) = nbr
    End If
'
    If (FC(1, nbr) = a) Then
        J = 1
    Else
        J = 2
    End If
'
    BF(3 - J, L) = Ptr
    a = FC(3 - J, nbr)
    Ptr = nbr
    If (Ptr <> bfi) Then GoTo 30
'
'600 format (1x,'new tetra: ',4i7)
'
'
'
End Sub
Private Sub VBFAC(PT#(), Ctr#(), Vcl#(), VM&(), BF&(), ByRef FC&(), _
    ByRef Topv&, ByVal Topnv&)
'
'******************************************************************************
'
'  VBFAC determines the boundary faces of a 3D triangulation.
'
'
'  Purpose:
'
'    Determine boundary faces of 3D triangulation visible
'    from point PT, given a starting visible boundary face.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, PT(1:3) - 3D point.
'
'    Input, CTR(1:3) - 3D point in interior of triangulation.
'
'    Input, VCL(1:3,1:*) - vertex coordinate list.
'
'    Input, VM(1:*) - indices of vertices of VCL being triangulated.
'
'    Input, BF(1:3,1:*) -  array of boundary face records; see DTRIS3.
'
'    Input/output, FC(1:7,1:*) - array of face records; see routine DTRIS3;
'    row 7 is used for links of 3 stacks in this routine.  On output,
'    FC(7,*) has been updated, so that only stack of visible boundary
'    faces remains.
'
'    Input/output, TOPV.  On input, index of FC of visible boundary face.
'    On output, index of top of stack of visible boundary faces
'
'    Input, TOPNV - index of top of stack of boundary faces already found
'    to be not visible from PT, or 0 for empty stack.
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
'
    Dim J&, K&, nbr&, op&, Ptr&, topn&, topt&, va&, vb&, vc&, IERR&
    Dim PT1#(1 To 3), PT2#(1 To 3), PT3#(1 To 3), PT4#(1 To 3) ' Vettori d' appoggio.
'
  ' TOPN is index of top of stack of non-visible boundary faces.
  ' TOPT is index of top of stack of boundary faces to be tested.
'
    topn = Topnv
    topt = 0
    FC(7, Topv) = 0
    K = -FC(5, Topv)
'
    For J = 1 To 3
        nbr = BF(J, K)
        If (FC(7, nbr) = -1) Then
            FC(7, nbr) = topt
            topt = nbr
        End If
    Next J
'
    Do While (topt <> 0)
        Ptr = topt
        topt = FC(7, Ptr)
        va = VM(FC(1, Ptr))
        vb = VM(FC(2, Ptr))
        vc = VM(FC(3, Ptr))
        'op = OPSIDE(Vcl(1, va), Vcl(1, vb), Vcl(1, vc), ctr, PT)
        PT1(1) = Vcl(1, va): PT1(2) = Vcl(2, va): PT1(3) = Vcl(3, va)
        PT2(1) = Vcl(1, vb): PT2(2) = Vcl(2, vb): PT2(3) = Vcl(3, vb)
        PT3(1) = Vcl(1, vc): PT3(2) = Vcl(2, vc): PT3(3) = Vcl(3, vc)
        op = OPSIDE(PT1(), PT2(), PT3(), Ctr(), PT())
'
        If (op = 2) Then
            IERR = 301
            Exit Sub
        End If
'
        If (op = 1) Then
            FC(7, Ptr) = Topv
            Topv = Ptr
            K = -FC(5, Ptr)
'
            For J = 1 To 3
                nbr = BF(J, K)
                If (FC(7, nbr) = -1) Then
                    FC(7, nbr) = topt
                    topt = nbr
                End If
            Next J
'
        Else
            FC(7, Ptr) = topn
            topn = Ptr
        End If
    Loop
'
    ' For boundary faces not visible from PT, set FC(7,*) = -1.
'
    Do While (topn <> 0)
        Ptr = topn
        topn = FC(7, Ptr)
        FC(7, Ptr) = -1
    Loop
'
'
'
End Sub
Private Function OPSIDE(a#(), b#(), c#(), d#(), e#()) As Long
'
'******************************************************************************
'
'  OPSIDE tests if points are on opposite sides of a triangular face.
'
'
'  Purpose:
'
'    Test if points D, E are on opposite sides of triangular
'    face with vertices A, B, C.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, A(1:3), B(1:3), C(1:3), D(1:3), E(1:3) - five 3D points.
'
'    Output, OPSIDE, the result of the test:
'    +1 if D, E on opposite sides;
'    -1 if on same side;
'     2 if D is coplanar with face ABC (ABCD is a degenerate tetrahedron);
'     0 if E is coplanar with face ABC
'
'
    Dim I&
    Dim ddp#, dmax#, edp#, emax#, nrml1#, nrml2#, nrml3#, tol#
    Dim ab#(1 To 3), ac#(1 To 3)
'
    tol = 100# * Epsilon
'
    'ab(1:3) = b(1:3) - a(1:3)
    'ac(1:3) = c(1:3) - a(1:3)
    For I = 1 To 3
        ab(I) = b(I) - a(I)
        ac(I) = c(I) - a(I)
    Next I
'
    emax = DMAX1(Abs(a(1)), Abs(a(2)), Abs(a(3)), Abs(b(1)), Abs(b(2)), _
                 Abs(b(3)), Abs(c(1)), Abs(c(2)), Abs(c(3)))
    dmax = DMAX1(emax, Abs(d(1)), Abs(d(2)), Abs(d(3)))
    nrml1 = ab(2) * ac(3) - ab(3) * ac(2)
    nrml2 = ab(3) * ac(1) - ab(1) * ac(3)
    nrml3 = ab(1) * ac(2) - ab(2) * ac(1)
    ddp = (d(1) - a(1)) * nrml1 + (d(2) - a(2)) * nrml2 _
        + (d(3) - a(3)) * nrml3
'
    If (Abs(ddp) <= tol * dmax) Then
        OPSIDE = 2
        Exit Function
    End If
'
    emax = DMAX1(emax, Abs(e(1)), Abs(e(2)), Abs(e(3)))
    edp = (e(1) - a(1)) * nrml1 + (e(2) - a(2)) * nrml2 _
        + (e(3) - a(3)) * nrml3
'
    If (Abs(edp) <= tol * emax) Then
        OPSIDE = 0
    ElseIf (ddp * edp < 0#) Then
        OPSIDE = 1
    Else
        OPSIDE = -1
    End If
'
'
'
End Function
Private Sub NWTHIN(ByVal I&, ByVal Ifac&, ByVal Ivrt&, ByVal NPT&, ByVal Sizht&, _
    ByRef Nfc&, ByVal Maxfc&, ByRef FC&(), ByRef HT&(), ByRef Ntetra&, _
    ByRef hdavfc&, ByRef Front&, ByRef Back&, ByRef IERR&)
'
'******************************************************************************
'
'  NWTHIN creates new tetrahedra after the insertion of an interior vertex.
'
'
'  Purpose:
'
'    Create new tetrahedra in 3D triangulation from the
'    insertion of vertex I in interior of tetrahedron.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, I - (local) index of next vertex inserted in triangulation.
'
'    Input, IFAC - face of tetrahedron containing vertex I is FC(*,IFAC).
'
'    Input, IVRT - 4 or 5 where 4th vertex of tetrahedron is FC(IVRT,IFAC).
'
'    Input, NPT - number of 3D points to be triangulated.
'
'    Input, SIZHT - size of hash table HT.
'
'    Input/output, NFC - number of positions used in FC array.
'
'    Input, MAXFC - maximum size available for FC array.
'
'    Input/output, FC(1:7,1:MAXFC) - array of face records; see routine DTRIS3.
'
'    Input/output, HT(0:SIZHT-1) - hash table using direct chaining.
'
'    Input/output, NTETRA - number of tetrahedra in triangulation.
'
'    Input/output, HDAVFC - head pointer to available FC records.
'
'    Output, FRONT, BACK - indices of front and back of queue of interior
'    faces ABC such that ABCI is a new tetrahedron.
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
'
    Dim a&, aa&, b&, Bb&, c&, cc&, d&, dd&, J&, Ind&
    ReDim Indx&(1 To 6)
    Const msglvl& = 0
'
    IERR = 0
    Front = 0
    Back = 0
    Ntetra = Ntetra + 3
    a = FC(1, Ifac)
    b = FC(2, Ifac)
    c = FC(3, Ifac)
    d = FC(Ivrt, Ifac)
'
    For J = 1 To 4
        If (J = 1) Then
            aa = a
            Bb = b
            cc = c
            dd = d
            Ind = Ifac
'
        Else
            If (J = 2) Then
                cc = d
                dd = c
            ElseIf (J = 3) Then
                Bb = c
                dd = b
            Else
                aa = b
                dd = a
            End If
'
            Ind = HTSRC(aa, Bb, cc, NPT, Sizht, FC(), HT())
'
            If (Ind <= 0) Then
                IERR = 300
                Exit Sub
            End If
'
        End If
'
        If (FC(4, Ind) = dd) Then
            FC(4, Ind) = I
        Else
            FC(5, Ind) = I
        End If
'
        If (FC(5, Ind) > 0) Then
            If (Front = 0) Then
                Front = Ind
            Else
                FC(7, Back) = Ind
            End If
            Back = Ind
        End If
'
        If (msglvl = 4) Then
            'write ( *,600) aa,bb,cc,i
        End If
    Next J
'
    If (Front <> 0) Then FC(7, Back) = 0
'
    For J = 1 To 6
        Call AVAILF(hdavfc, Nfc, Maxfc, FC(), Indx(J), IERR)
        If (IERR <> 0) Then Exit Sub
    Next J
'
    Call HTINS(Indx(1), a, b, I, c, d, NPT, Sizht, FC(), HT())
    Call HTINS(Indx(2), a, c, I, b, d, NPT, Sizht, FC(), HT())
    Call HTINS(Indx(3), a, d, I, b, c, NPT, Sizht, FC(), HT())
    Call HTINS(Indx(4), b, c, I, a, d, NPT, Sizht, FC(), HT())
    Call HTINS(Indx(5), b, d, I, a, c, NPT, Sizht, FC(), HT())
    Call HTINS(Indx(6), c, d, I, a, b, NPT, Sizht, FC(), HT())
'
'600 format (1x,'new tetra: ',4i7)
'
'
'
End Sub
Private Sub WALKT3(PT#(), ByVal N&, ByVal P&, ByVal Ntetra&, Vcl#(), VM&(), _
    FC&(), HT&(), ByRef Ifac&, ByRef Ivrt&, ByRef IERR&)
'
'******************************************************************************
'
'  WALKT3 walks through a 3D triangulation searching for a point.
'
'
'  Purpose:
'
'    Walk through neighboring tetrahedra of 3D (Delaunay)
'    triangulation until a tetrahedron is found containing point PT
'    or PT is found to be outside the convex hull. Search is
'    guaranteed to terminate for a Delaunay triangulation, else a
'    cycle may occur.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, PT(1:3) - 3D point.
'
'    Input, N - upper bound on vertex indices or size of VM.
'
'    Input, P - size of hash table.
'
'    Input, NTETRA - number of tetrahedra in triangulation; used to detect
'    cycle.
'
'    Input, VCL(1:3,1:*) - vertex coordinate list.
'
'    Input, VM(1:N) - vertex mapping list (maps from local indices used in
'    FC to indices of VCL).
'
'    Input, FC(1:7,1:*) - array of face records; see routine DTRIS3.
'
'    Input, HT(0:P-1) - hash table using direct chaining.
'
'    Input/output, IFAC.  On input, index of face of FC to begin search at.
'    On output, index of FC indicating tetrahedron or face containing PT.
'
'    Input/output, IVRT.  On input, 4 or 5 to indicate tetrahedron to
'    begin search at.  On output, 4 or 5 to indicate that FC(IVRT,IFAC) is
'    4th vertex of tetrahedron containing PT in its interior; 6 if PT lies
'    in interior of face FC(*,IFAC); 0 if PT lies outside
'    convex hull on other side of face FC(*,IFAC); 1, 2, or 3
'    if PT lies on interior of edge of face from vertices
'    FC(IVRT,IFAC) to FC(IVRT mod 3 + 1,IFAC); -1, -2, or -3
'    if PT is (nearly) vertex FC(-IVRT,IFAC).
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
'
    Dim J&, K&, a&, aa&, b&, Bb&, c&, cnt&, d&, va&, vb&, vc&, Vd&
    Dim Degen As Boolean, Zero(4) As Boolean
    Dim T#(1 To 4)
    Dim PT1#(1 To 3), PT2#(1 To 3), PT3#(1 To 3), PT4#(1 To 3) ' Vettori d' appoggio.
'
    IERR = 0
    cnt = 0
'
10  'continue
'
    If (FC(Ivrt, Ifac) <= 0) Then
        Ivrt = 0
        Exit Sub
    End If
'
    cnt = cnt + 1
'
    If (cnt > Ntetra) Then
        IERR = 307
        Exit Sub
        End If
'
    a = FC(1, Ifac)
    b = FC(2, Ifac)
    c = FC(3, Ifac)
    d = FC(Ivrt, Ifac)
    va = VM(a)
    vb = VM(b)
    vc = VM(c)
    Vd = VM(d)
    'Call BARYTH(Vcl(1, va), Vcl(1, vb), Vcl(1, vc), Vcl(1, Vd), PT, T, degen)
    PT1(1) = Vcl(1, va): PT1(2) = Vcl(2, va): PT1(3) = Vcl(3, va)
    PT2(1) = Vcl(1, vb): PT2(2) = Vcl(2, vb): PT2(3) = Vcl(3, vb)
    PT3(1) = Vcl(1, vc): PT3(2) = Vcl(2, vc): PT3(3) = Vcl(3, vc)
    PT4(1) = Vcl(1, Vd): PT4(2) = Vcl(2, Vd): PT4(3) = Vcl(3, Vd)
    Call BARYTH(PT1(), PT2(), PT3(), PT4(), PT(), T(), Degen)
'
    If (Degen) Then
        IERR = 301
        Exit Sub
    End If
'
    If (T(1) > 0# And T(2) > 0# _
    And T(3) > 0# And T(4) > 0#) Then
        Exit Sub
'
    ElseIf (T(4) < 0#) Then
        Ivrt = 9 - Ivrt
'
    ElseIf (T(1) < 0#) Then
        Ifac = HTSRC(b, c, d, N, P, FC(), HT())
        If (Ifac <= 0) Then GoTo 30
'
        If (FC(4, Ifac) = a) Then
            Ivrt = 5
        Else
            Ivrt = 4
        End If
'
    ElseIf (T(2) < 0#) Then
        Ifac = HTSRC(a, c, d, N, P, FC(), HT())
        If (Ifac <= 0) Then GoTo 30
'
        If (FC(4, Ifac) = b) Then
            Ivrt = 5
        Else
            Ivrt = 4
        End If
'
    ElseIf (T(3) < 0#) Then
        Ifac = HTSRC(a, b, d, N, P, FC(), HT())
'
        If (Ifac <= 0) Then GoTo 30
'
        If (FC(4, Ifac) = c) Then
            Ivrt = 5
        Else
            Ivrt = 4
        End If
'
    Else
        ' All T(J) >= 0.0D0 and at least one T(J) == 0.0D0
'
        K = 0
'
        For J = 1 To 4
            Zero(J) = (T(J) = 0#)
            If (Zero(J)) Then K = K + 1
        Next J
'
        If (K = 1) Then
            Ivrt = 6
'
            If (Zero(1)) Then
                Ifac = HTSRC(b, c, d, N, P, FC(), HT())
                If (Ifac <= 0) Then GoTo 30
            ElseIf (Zero(2)) Then
                Ifac = HTSRC(a, c, d, N, P, FC(), HT())
                If (Ifac <= 0) Then GoTo 30
            ElseIf (Zero(3)) Then
                Ifac = HTSRC(a, b, d, N, P, FC(), HT())
                If (Ifac <= 0) Then GoTo 30
            End If
'
        ElseIf (K = 2) Then
            If (Zero(4)) Then
                If (Zero(3)) Then
                    Ivrt = 1
                ElseIf (Zero(1)) Then
                    Ivrt = 2
                Else
                    Ivrt = 3
                End If
'
            Else
                If (Zero(3)) Then
                    Ifac = HTSRC(a, b, d, N, P, FC(), HT())
                    If (Ifac <= 0) Then GoTo 30
'
                    If (Zero(2)) Then
                        aa = a
                    Else
                        aa = b
                    End If
'
                Else
                    Ifac = HTSRC(a, c, d, N, P, FC(), HT())
                    If (Ifac <= 0) Then GoTo 30
                    aa = c
                End If
'
                Bb = d
'
                If (aa > Bb) Then
                    J = aa
                    aa = Bb
                    Bb = J
                End If
'
                If (FC(1, Ifac) = aa) Then
                    If (FC(2, Ifac) = Bb) Then
                        Ivrt = 1
                    Else
                        Ivrt = 3
                    End If
'
                Else
                    Ivrt = 2
                End If
            End If
'
        Else
'
            ' K == 3
            If (Not Zero(1)) Then
                Ivrt = -1
'
            ElseIf (Not Zero(2)) Then
                Ivrt = -2
'
            ElseIf (Not Zero(3)) Then
                Ivrt = -3
'
            Else
                Ifac = HTSRC(a, b, d, N, P, FC(), HT())
                If (Ifac <= 0) Then GoTo 30
'
                If (FC(1, Ifac) = d) Then
                    Ivrt = -1
                ElseIf (FC(2, Ifac) = d) Then
                    Ivrt = -2
                Else
                    Ivrt = -3
                End If
            End If
        End If
'
        Exit Sub
'
    End If
'
    GoTo 10
'
30  'continue
'
IERR = 300
'
'
'
End Sub
Private Sub HTINS(ByVal Ind&, ByVal a&, ByVal b&, ByVal c&, ByVal d&, ByVal e&, _
    ByVal N&, ByVal P&, ByRef FC&(), ByRef HT&())
'
'******************************************************************************
'
'  HTINS inserts a record into the hash table.
'
'
'  Purpose:
'
'    Insert record FC(1:7,IND) containing A,B,C,D,E,HTLINK,-1
'    into hash table HT.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, IND - index of FC array.
'
'    Input, A, B, C, D, E - first 5 fields of FC record (or column).
'
'    Input, N - upper bound on vertex indices.
'
'    Input, P - size of hash table.
'
'    Input/output, FC(1:7,1:*) - array of face records; see routine DTRIS3.
'
'    Input/output, HT(0:P-1) - hash table using direct chaining
'
'
    Dim aa&, Bb&, cc&, K&
'
    aa = a
    Bb = b
    cc = c
    Call ORDER3(aa, Bb, cc)
    'k = mod(aa*n + bb, p)
    'k = mod(k*n + cc, p)
    K = (aa * N + Bb) Mod P
    K = (K * N + cc) Mod P
    FC(1, Ind) = aa
    FC(2, Ind) = Bb
    FC(3, Ind) = cc
    FC(4, Ind) = d
    FC(5, Ind) = e
    FC(6, Ind) = HT(K)
    FC(7, Ind) = -1
    HT(K) = Ind
'
'
'
End Sub
Private Sub FRSTET(ByVal Shift As Boolean, ByVal NV&, Vcl#(), MAP&(), _
    ByRef I3&, ByRef I4&, ByRef IERR&)
'
'******************************************************************************
'
'  FRSTET shifts vertices so the first 4 vertices are in general position in 3D.
'
'
'  Purpose:
'
'    Shift or swap vertices if necessary so first 3 vertices
'    (according to MAP) are not collinear and first 4 vertices are
'    not coplanar (so that first tetrahedron is valid).
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, SHIFT - if .TRUE., MAP(3), MAP(4) may be updated due to shift,
'    else they may be updated due to swaps; in former case,
'    it is assumed MAP gives vertices in lexicographic order.
'
'    Input, NV - number of vertices.
'
'    Input, VCL(1:3,1:*) - vertex coordinate list.
'
'    Input/output, MAP(1:NV), on input, contains vertex indices of VCL.
'    On output, shifted or 2 swaps applied if necessary so that vertices
'    indexed by MAP(1), MAP(2), MAP(3), MAP(4) not coplanar.
'
'    Output, I3, I4 - the indices such that MAP_in(I3) = MAP_out(3) and
'    MAP_in(I4) = MAP_out(4).
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
'
    Dim cmax#, cp1#, cp2#, cp3#, dmax#, dotp#, dV2#(3), dvk#(3), dvl#(3), tol#
    Dim I&, K&, L&, M&, M1&, M2&
'
    ' First check that consecutive vertices are not identical.
'
    IERR = 0
    tol = 100# * Epsilon
'
    If (Shift) Then
        L = NV - 1
    Else
        L = 1
    End If
'
    M1 = MAP(1)
'
    For I = 1 To L
        M = M1
        M1 = MAP(I + 1)
        For K = 1 To 3
            cmax = DMAX1(Abs(Vcl(K, M)), Abs(Vcl(K, M1)))
            If (Abs(Vcl(K, M) - Vcl(K, M1)) > tol * cmax And cmax > tol) Then
                GoTo 20
            End If
        Next K
'
        IERR = 302
        Exit Sub
'
20      'continue
    Next I
'
    ' Find index K = I3 and L = I4.
'
    M1 = MAP(1)
    M2 = MAP(2)
    dV2(1) = Vcl(1, M2) - Vcl(1, M1)
    dV2(2) = Vcl(2, M2) - Vcl(2, M1)
    dV2(3) = Vcl(3, M2) - Vcl(3, M1)
    cmax = DMAX1(Abs(Vcl(1, M1)), Abs(Vcl(2, M1)), Abs(Vcl(3, M1)), _
           Abs(Vcl(1, M2)), Abs(Vcl(2, M2)), Abs(Vcl(3, M2)))
    K = 2
'
30  'continue
'
    K = K + 1
'
    If (K > NV) Then
        IERR = 303
        Exit Sub
    End If
'
    M = MAP(K)
    dvk(1) = Vcl(1, M) - Vcl(1, M1)
    dvk(2) = Vcl(2, M) - Vcl(2, M1)
    dvk(3) = Vcl(3, M) - Vcl(3, M1)
    dmax = DMAX1(cmax, Abs(Vcl(1, M)), Abs(Vcl(2, M)), Abs(Vcl(3, M)))
    cp1 = dV2(2) * dvk(3) - dV2(3) * dvk(2)
    cp2 = dV2(3) * dvk(1) - dV2(1) * dvk(3)
    cp3 = dV2(1) * dvk(2) - dV2(2) * dvk(1)
'
    If (DMAX1(Abs(cp1), Abs(cp2), Abs(cp3)) <= tol * dmax) Then
        GoTo 30
    End If
'
    cmax = dmax
    L = K
'
40  'continue
'
    L = L + 1
'
    If (L > NV) Then
        IERR = 304
        Exit Sub
    End If
'
    M = MAP(L)
    dvl(1) = Vcl(1, M) - Vcl(1, M1)
    dvl(2) = Vcl(2, M) - Vcl(2, M1)
    dvl(3) = Vcl(3, M) - Vcl(3, M1)
    dmax = DMAX1(cmax, Abs(Vcl(1, M)), Abs(Vcl(2, M)), Abs(Vcl(3, M)))
    dotp = dvl(1) * cp1 + dvl(2) * cp2 + dvl(3) * cp3
'
    If (Abs(dotp) <= tol * dmax) Then GoTo 40
'
    ' Shift or swap elements of MAP if necessary.
'
    If (Shift) Then
        If (K > 3) Then
            M1 = MAP(K)
        End If
'
        If (L > 4) Then
            M2 = MAP(L)
            For I = L To K + 2 Step -1
                MAP(I) = MAP(I - 1)
            Next I
            For I = K + 1 To 5 Step -1
                MAP(I) = MAP(I - 2)
            Next I
            MAP(4) = M2
        End If
'
        If (K > 3) Then MAP(3) = M1
'
    Else
        If (K > 3) Then
            M = MAP(3)
            MAP(3) = MAP(K)
            MAP(K) = M
        End If
'
        If (L > 4) Then
            M = MAP(4)
            MAP(4) = MAP(L)
            MAP(L) = M
        End If
'
    End If
'
    I3 = K
    I4 = L
'
'
'
End Sub
Private Sub ORDER3(ByRef I&, ByRef J&, ByRef K&)
'
'******************************************************************************
'
'  ORDER3 reorders 3 integers into ascending order.
'
'
'  Purpose:
'
'    Order I, J, K so that I <= J <= K.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input/output, integer I, J, K, on output are sorted into
'    nondecreasing order.
'
'
    Dim T&
'
    If (J < I) Then
        If (K < J) Then
            Call I_SWAP(I, K)
        ElseIf (K < I) Then
            T = I
            I = J
            J = K
            K = T
        Else
            Call I_SWAP(I, J)
        End If
'
    Else
        If (K < I) Then
            T = I
            I = K
            K = J
            J = T
        ElseIf (K < J) Then
            Call I_SWAP(J, K)
        End If
    End If
'
'
'
End Sub
Private Sub NWTHFC(ByVal I&, ByVal Ifac&, ByVal NPT&, ByVal Sizht&, _
    ByRef Nbf&, ByRef Nfc&, ByVal Maxbf&, ByVal Maxfc&, _
    ByRef BF&(), ByRef FC&(), ByRef HT&(), ByRef Ntetra&, _
    ByRef hdavbf&, ByRef hdavfc&, ByRef Front&, ByRef Back&, ByRef IERR&)
'
'******************************************************************************
'
'  NWTHFC creates new tetrahedra after the insertion of a new face vertex.
'
'
'  Purpose:
'
'    Create new tetrahedra in 3D triangulation from the
'    insertion of vertex I on face FC(*,IFAC).
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, I - (local) index of next vertex inserted in triangulation;
'    it is assumed I is largest index so far.
'
'    Input, IFAC - face containing vertex I is FC(*,IFAC).
'
'    Input, NPT - number of 3D points to be triangulated.
'
'    Input, SIZHT - size of hash table HT.
'
'    Input/output, NBF - number of positions used in BF array.
'
'    Input/output, NFC - number of positions used in FC array.
'
'    Input, MAXBF - maximum size available for BF array.
'
'    Input, MAXFC - maximum size available for FC array.
'
'    Input/output, BF(1:3,1:MAXBF) -  array of boundary face records;
'    see DTRIS3.
'
'    Input/output, FC(1:7,1:MAXFC) - array of face records; see routine DTRIS3.
'
'    Input/output, HT(0:SIZHT-1) - hash table using direct chaining.
'
'    Input/output, NTETRA - number of tetrahedra in triangulation.
'
'    Input/output, HDAVBF - head pointer to available BF records.
'
'    Input/output, HDAVFC - head pointer to available FC records.
'
'    Output, FRONT, BACK - indices of front and back of queue of interior
'    faces ABC such that ABCI is a new tetrahedron.
'
'    Output, integer IERR, error flag, which is zero unless an error occurred.
'
'
    Dim a&, aa&, b&, Bb&, bf1&, bf2&, c&, cc&, d&, e&
    Dim Ind&, ind2&, Iv&, J&, nbrac&, nbrbc&, NV&
    Dim bface As Boolean
    Const msglvl& = 0
'
    IERR = 0
    Front = 0
    Back = 0
    NV = 5
    bface = (FC(5, Ifac) <= 0)
    If (bface) Then NV = 4
    a = FC(1, Ifac)
    b = FC(2, Ifac)
    c = FC(3, Ifac)
'
    For Iv = 4 To NV
        Ntetra = Ntetra + 2
        d = FC(Iv, Ifac)
'
        For J = 1 To 3
            If (J = 1) Then
                aa = a
                Bb = b
                cc = c
            ElseIf (J = 2) Then
                aa = b
                Bb = c
                cc = a
            Else
                aa = c
                Bb = a
                cc = b
            End If
'
            Call AVAILF(hdavfc, Nfc, Maxfc, FC(), Ind, IERR)
            If (IERR <> 0) Then Exit Sub
'
            Call HTINS(Ind, aa, d, I, Bb, cc, NPT, Sizht, FC(), HT())
            Ind = HTSRC(aa, Bb, d, NPT, Sizht, FC(), HT())
'
            If (Ind <= 0) Then
                IERR = 300
                Exit Sub
            End If
'
            If (FC(4, Ind) = cc) Then
                FC(4, Ind) = I
            Else
                FC(5, Ind) = I
            End If
'
            If (FC(5, Ind) > 0) Then
                If (Front = 0) Then
                    Front = Ind
                Else
                    FC(7, Back) = Ind
                End If
                Back = Ind
            End If
'
            If (msglvl = 4) Then
                'write ( *,600) aa,bb,d,i
            End If
        Next J
    Next Iv
'
    If (Front <> 0) Then FC(7, Back) = 0
'
    Call AVAILF(hdavfc, Nfc, Maxfc, FC(), Ind, IERR)
    Call AVAILF(hdavfc, Nfc, Maxfc, FC(), ind2, IERR)
    If (IERR <> 0) Then Exit Sub
'
    If (bface) Then
        e = FC()(5, Ifac)
    Else
        e = FC(4, Ifac)
    End If
'
    Call HTDEL(Ifac, NPT, Sizht, FC(), HT())
    Call HTINS(Ifac, a, b, I, d, e, NPT, Sizht, FC(), HT())
    Call HTINS(Ind, a, c, I, d, e, NPT, Sizht, FC(), HT())
    Call HTINS(ind2, b, c, I, d, e, NPT, Sizht, FC(), HT())
'
    If (bface) Then
        e = -e
'
        If (hdavbf <> 0) Then
            bf1 = hdavbf
            hdavbf = -BF(1, hdavbf)
            If (hdavbf <> 0) Then
                bf2 = hdavbf
                hdavbf = -BF(1, hdavbf)
            Else
                Nbf = Nbf + 1
                bf2 = Nbf
            End If
'
        Else
            Nbf = Nbf + 2
            bf1 = Nbf - 1
            bf2 = Nbf
        End If
'
        If (Nbf > Maxbf) Then
            Nbf = Maxbf
            IERR = 12
            Exit Sub
        End If
'
        FC(5, Ind) = -bf1
        FC(5, ind2) = -bf2
        nbrac = BF(2, e)
        nbrbc = BF(1, e)
        BF(1, e) = ind2
        BF(2, e) = Ind
        BF(1, bf1) = ind2
        BF(2, bf1) = Ifac
        BF(3, bf1) = nbrac
        BF(1, bf2) = Ind
        BF(2, bf2) = Ifac
        BF(3, bf2) = nbrbc
        J = -FC(5, nbrac)
'
        If (BF(1, J) = Ifac) Then
            BF(1, J) = Ind
        ElseIf (BF(2, J) = Ifac) Then
            BF(2, J) = Ind
        Else
            BF(3, J) = Ind
        End If
'
        J = -FC(5, nbrbc)
'
        If (BF(1, J) = Ifac) Then
            BF(1, J) = ind2
        ElseIf (BF(2, J) = Ifac) Then
            BF(2, J) = ind2
        Else
            BF(3, J) = ind2
        End If
    End If
'
'600 format (1x,'new tetra: ',4i7)
'
'
'
End Sub
Private Sub zCodiciErrore()
'
'
'   Codici di errore originali della libreria GEOMPACK3:
'
'   Error codes: Error codes IERR from 1 to 99 indicate not enough space in some array.
'   Error codes IERR >= 100 mean that the input specification is incorrect,
'    there is a program bug, or the tolerance TOL is too small or large (try
'   calling INITCB with a different TOLIN value).
'
'   IERR    DESCRIPTION OF ERROR
'     1  -  not enough space in EDGE array for routine EDGHT
'     2  -  not enough space in HOLV array for routine DSMCPR or DSPGDC
'     3  -  not enough space in 2-D VCL array
'     4  -  not enough space in HVL array
'     5  -  not enough space in PVL, IANG arrays
'     6  -  not enough space in IWK array
'     7  -  not enough space in WK array
'     8  -  not enough space in STACK array for routine SWAPEC, DTRIS2, or DTRIW2
'     9  -  not enough space in TIL array
'    10  -  not enough space in CWALK array for routine INTTRI
'    11  -  not enough space in FC array for 3-D
'    12  -  not enough space in BF array for 3-D
'    13  -  not enough space in SVCL, SFVL arrays for routine SHRNK3
'    14  -  not enough space in 3-D VCL array
'    15  -  not enough space in FVL, EANG arrays for polyhedral decomposition
'    16  -  not enough space in FACEP, NRML arrays
'    17  -  not enough space in PFL array
'    18  -  not enough space in HFL array
'    19  -  not enough space in BTL array
'    20  -  not enough space in VM array
'    21  -  not enough space in HT array
'    22  -  not enough space in FC array for K-D
'    23  -  not enough space in BF array for K-D
'   100  -  higher primes need to be added to routine PRIME
'   200  -  abnormal return in routine DIAM2
'   201  -  abnormal return in routine WIDTH2
'   202  -  parallel lines in routine SHRNK2
'   205  -  horizontal ray to right does not intersect polygon in routine ROTIPG
'   206  -  out-of-range index when popping points from stack in routine VPRGHT
'   207  -  out-of-range index when scanning vertices in routine VPSCNA
'   208  -  out-of-range index when scanning vertices in routine VPSCNB
'   209  -  out-of-range index when scanning vertices in routine VPSCNC
'   210  -  out-of-range index when scanning vertices in routine VPSCND
'   212  -  singular matrix in routine VORNBR
'   215  -  unmatched separator interface edge determined by routine DSPGDC
'   216  -  all edges of an outer boundary curve are specified as hole
'           edges in input to routine DSPGDC
'   218  -  cannot find subregion above top vertex of hole in routine JNHOLE
'   219  -  angle at hole vertex in modified region is too far from PI due
'           to use of relative tolerance in routine JNHOLE
'   222  -  separator not found in routine MFDEC2 (not likely to occur if
'           ANGSPC <= 30 degrees and ANGTOL <= 20 degrees)
'   224  -  2 vertices with identical coordinates (in floating point arithmetic)
'           detected in routine DTRIS2 or DTRIW2
'   225  -  all vertices input to DTRIS2 or DTRIW2 are collinear (in fl pt arith)
'   226  -  cycle detected in walk by routine WALKT2
'   230  -  invalid (CW-oriented) triangle created in routine TMERGE
'   231  -  unsuccessful search in routine FNDTRI
'   300  -  unsuccessful search by routine HTSRC
'   301  -  degenerate tetrahedron detected by routine BARYTH, CCSPH, or OPSIDE
'   302  -  2 vertices with identical coordinates (in floating point arithmetic)
'           detected in routine FRSTET, DTRIS3, DTRIW3, or ITRIS3
'   303  -  all vertices input to DTRIS3 or DTRIW3 are collinear (in fl pt arith)
'   304  -  all vertices input to DTRIS3 or DTRIW3 are coplanar (in fl pt arith)
'   305  -  invalid configuration (due to tolerance) detected by routine SWAPES
'   306  -  no visible boundary face found in routine DTRIS3 or ITRIS3
'   307  -  cycle detected in walk by routine WALKT3
'   308  -  nontransformable or nonexistent face at top of stack in routine SWAPTF
'   309  -  ALPHA(4) = 0 occurs after calling BARYTH in SWAPES, SWAPMU, etc.;
'           TOL must be decreased
'   310  -  unmatched edge determined by routine DSCPH
'   313  -  incorrect value for variable T in routine XPGHPL
'   314  -  unmatched edge in shrunken polyhedron in routine SHRNK3
'   315  -  2 "identical" vertices with unequal coordinates in routine SHRNK3
'   318  -  vertex ITOP not found in FVL array in routine INTMVG
'   321  -  face oriented same way twice in routine DSPHDC or DSPHFH
'   322  -  unmatched edge determined by routine DSPHDC, DSPHFH, or DSPHIH
'   325  -  unsuccessful search for next edge of cut polygon in routine CUTFAC
'   326  -  unsuccessful search for intersecting edge of face FL in CUTFAC
'   327  -  at least one reflex edge has not been resolved by routine RESEDG
'   328  -  value of constant MAXEV must be increased in routine CUTFAC
'   330  -  value of MAXSV must be increased before calling SHRNK3 in TRIPR3
'   331  -  invalid vertex detected from walk or linear search in routine BCDTRI
'   334  -  unsuccessful search for next edge of cut polygon in routine SEPFAC
'   335  -  unsuccessful search for intersecting edge of face FL in SEPFAC
'   336  -  separator face not found in routine MFDEC3
'   340  -  holes are not in nondecreasing face index order in routine DSPHFH
'   341  -  hole on non-boundary face of polyhedral region detected by DSPHFH
'   342  -  face containing non-coplanar vertices detected by DSPHFH
'   344  -  cannot find subregion above top (or below bottom) vertex of hole in
'           routine SPDECF
'   346  -  hole polyhedron not connected to outer boundary by routine DSPHIH
'   347  -  starting edge not found by routine RESHOL (because PT not in polyh)
'   348  -  starting edge not found by routine RESHOL (because ray meets vertex)
'   349  -  unsuccessful search for intersecting edge of face FMIN in RESHOL
'   400  -  unsuccessful search by routine HTSRCK or HTSDLK
'   401  -  degenerate simplex detected by routine BARYCK
'   402  -  2 vertices with identical coordinates (in floating point arithmetic)
'           detected in routine FRSMPX, DTRISK, DTRIWK, or DTRIMK
'   403  -  all vertices input to DTRISK, DTRIWK, or DTRIMK are in same hyperplane
'           (in floating point arithmetic)
'   405  -  invalid configuration (due to tolerance) detected by routine SWAPHS
'   406  -  no visible boundary face found in routine DTRISK
'   407  -  cycle detected in walk by routine WALKTK
'
'
'
End Sub
Private Function HTSRC(ByVal a&, ByVal b&, ByVal c&, ByVal N&, ByVal P&, _
    FC&(), HT&()) As Long
'
'******************************************************************************
'
'  HTSRC searches for a record in the hash table.
'
'
'  Purpose:
'
'    Search for record FC(1:7,IND) containing key A,B,C
'    in hash table HT.
'
'  Author:
'
'    Barry Joe,
'    Department of Computing Science,
'    University of Alberta,
'    Edmonton, Alberta, Canada  T6G 2H1
'    Email: barry@cs.ualberta.ca
'
'  Parameters:
'
'    Input, A,B,C - first 3 fields of FC record (in any order).
'
'    Input, N - upper bound on vertex indices.
'
'    Input, P - size of hash table.
'
'    Input, FC(1:7,1:*) - array of face records; see routine DTRIS3.
'
'    Input, HT(0:P-1) - hash table using direct chaining.
'
'    Output, HTSRC - index of FC record with key A,B,C if found,
'    or 0 if not found.
'
'
    Dim aa&, Bb&, cc&, Ind&, K&
'
    aa = a
    Bb = b
    cc = c
    Call ORDER3(aa, Bb, cc)
    K = (aa * N + Bb) Mod P
    K = (K * N + cc) Mod P
    Ind = HT(K)
'
    Do
        If (Ind = 0) Then
            Exit Do
        End If
'
        If (FC(1, Ind) = aa And FC(2, Ind) = Bb And FC(3, Ind) = cc) Then
            Exit Do
        End If
'
        Ind = FC(6, Ind)

    Loop
'
    HTSRC = Ind
'
'
'
End Function

